# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines

import abc
import collections
import enum
import ipaddress

import attr

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.types import convert_url, convert_value_to_object

from cryptoparser.common.base import (
    Serializable,
    StringEnumCaseInsensitiveParsable,
    StringEnumParsable,
    VariantParsable
)
from cryptoparser.common.exception import InvalidType

from cryptoparser.common.field import (
    FieldValueComponentParsable,
    FieldValueComponentParsableOptional,
    FieldValueComponentNumber,
    FieldValueComponentPercent,
    FieldValueComponentString,
    FieldValueComponentUrl,
    FieldValueSingleBase,
    FieldValueStringEnumParams,
    FieldsSemicolonSeparated,
    NameValuePairListSemicolonSeparated,
    NameValuePair,
)
from cryptoparser.common.parse import ComposerText, ParsableBase, ParserText


class DmarcAlignment(StringEnumCaseInsensitiveParsable, enum.Enum):
    RELAXED = FieldValueStringEnumParams(
        code='r',
        human_readable_name='Relaxed',
    )
    STRICT = FieldValueStringEnumParams(
        code='s',
        human_readable_name='Strict',
    )


class DmarcIdentifierAlignmentBase(FieldValueComponentParsable):
    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    def _get_value_class(cls):
        return DmarcAlignment


class DnsRecordTxtValueDmarcValueIdentifierAlignmentDkim(DmarcIdentifierAlignmentBase):
    @classmethod
    def get_canonical_name(cls):
        return 'adkim'


class DnsRecordTxtValueDmarcValueIdentifierAlignmentAspf(DmarcIdentifierAlignmentBase):
    @classmethod
    def get_canonical_name(cls):
        return 'aspf'


class DmarcPolicyOption(StringEnumCaseInsensitiveParsable, enum.Enum):
    NONE = FieldValueStringEnumParams(
        code='none'
    )
    QUARANTINE = FieldValueStringEnumParams(
        code='quarantine'
    )
    REJECT = FieldValueStringEnumParams(
        code='reject'
    )


class DnsRecordTxtValueDmarcValuePolicy(FieldValueComponentParsable):
    @classmethod
    def get_canonical_name(cls):
        return 'p'

    @classmethod
    def _get_value_class(cls):
        return DmarcPolicyOption


class DnsRecordTxtValueDmarcValueSubdomainPolicy(FieldValueComponentParsableOptional):
    @classmethod
    def get_canonical_name(cls):
        return 'sp'

    @classmethod
    def _get_value_class(cls):
        return DmarcPolicyOption


class DmarcPolicyVersion(StringEnumParsable, enum.Enum):
    DMARC1 = FieldValueStringEnumParams(
        code='DMARC1',
        human_readable_name='DMARC1',
    )


class DnsRecordTxtValueDmarcValueVersion(FieldValueComponentParsable):
    @classmethod
    def get_canonical_name(cls):
        return 'v'

    @classmethod
    def _get_value_class(cls):
        return DmarcPolicyVersion


class DmarcFailureReportingOption(StringEnumCaseInsensitiveParsable, enum.Enum):
    ALL_FAILURE = FieldValueStringEnumParams(
        code='0',
        human_readable_name='All Failure',
    )
    ANY_FAILURE = FieldValueStringEnumParams(
        code='1',
        human_readable_name='Any Failure',
    )
    DKIM_FAILURE = FieldValueStringEnumParams(
        code='d',
        human_readable_name='DKIM Failure',
    )
    SPF_FAILURE = FieldValueStringEnumParams(
        code='s',
        human_readable_name='SPF Failure',
    )


class DmarcValueFailureOption(FieldValueComponentParsable):
    @classmethod
    def get_canonical_name(cls):
        return 'fo'

    @classmethod
    def _get_value_class(cls):
        return DmarcFailureReportingOption


class DnsRecordTxtValueDmarcValuePercent(FieldValueComponentPercent):
    @classmethod
    def get_canonical_name(cls):
        return 'pct'


@attr.s
class DmarcReportingInterval(FieldValueComponentNumber):
    def __attrs_post_init__(self):
        if self.value < 0 or self.value >= 2 ** 32:
            raise InvalidValue(self.value, type(self), 'value')

    @classmethod
    def get_canonical_name(cls):
        return 'ri'


class DmarcFailureReportingFormat(StringEnumCaseInsensitiveParsable, enum.Enum):
    AUTHENTICATION_FAILURE_REPORTING_FORMAT = FieldValueStringEnumParams(
        code='afrf',
        human_readable_name='Authentication Failure Reporting Format (AFRF)',
    )


class DmarcReportingFormat(FieldValueComponentParsable):
    @classmethod
    def get_canonical_name(cls):
        return 'rf'

    @classmethod
    def _get_value_class(cls):
        return DmarcFailureReportingFormat


class DnsRecordTxtValueDmarcValueReportingUrlAggregated(FieldValueComponentUrl):
    @classmethod
    def get_canonical_name(cls):
        return 'rua'


class DnsRecordTxtValueDmarcValueReportingUrlFailure(FieldValueComponentUrl):
    @classmethod
    def get_canonical_name(cls):
        return 'ruf'


@attr.s
class DnsRecordTxtValueDmarc(FieldsSemicolonSeparated):  # pylint: disable=too-many-instance-attributes
    version = attr.ib(
        converter=DnsRecordTxtValueDmarcValueVersion.convert,
        validator=attr.validators.instance_of(DnsRecordTxtValueDmarcValueVersion)
    )
    policy = attr.ib(
        converter=DnsRecordTxtValueDmarcValuePolicy.convert,
        validator=attr.validators.instance_of(DnsRecordTxtValueDmarcValuePolicy)
    )
    alignment_dkim = attr.ib(
        default=DmarcAlignment.RELAXED,
        converter=DnsRecordTxtValueDmarcValueIdentifierAlignmentDkim.convert,
        validator=attr.validators.instance_of(DnsRecordTxtValueDmarcValueIdentifierAlignmentDkim),
        metadata={'human_readable_name': 'DKIM Alignment'},
    )
    alignment_aspf = attr.ib(
        default=DmarcAlignment.RELAXED,
        converter=DnsRecordTxtValueDmarcValueIdentifierAlignmentAspf.convert,
        validator=attr.validators.instance_of(DnsRecordTxtValueDmarcValueIdentifierAlignmentAspf),
        metadata={'human_readable_name': 'ASPF Alignment'},
    )
    failure_option = attr.ib(
        default=DmarcFailureReportingOption.ALL_FAILURE,
        converter=DmarcValueFailureOption.convert,
        validator=attr.validators.instance_of(DmarcValueFailureOption),
    )
    percent = attr.ib(
        default=100,
        converter=DnsRecordTxtValueDmarcValuePercent.convert,
        validator=attr.validators.instance_of(DnsRecordTxtValueDmarcValuePercent),
    )
    reporting_url_aggregated = attr.ib(
        default=None,
        converter=convert_value_to_object(DnsRecordTxtValueDmarcValueReportingUrlAggregated, convert_url()),
        validator=attr.validators.optional(
            attr.validators.instance_of(DnsRecordTxtValueDmarcValueReportingUrlAggregated)
        ),
        metadata={'human_readable_name': 'Aggregated Reporting URL'},
    )
    reporting_url_failure = attr.ib(
        default=None,
        converter=convert_value_to_object(DnsRecordTxtValueDmarcValueReportingUrlFailure, convert_url()),
        validator=attr.validators.optional(attr.validators.instance_of(DnsRecordTxtValueDmarcValueReportingUrlFailure)),
        metadata={'human_readable_name': 'Failure Reporting URL'},
    )
    reporting_format = attr.ib(
        default=DmarcFailureReportingFormat.AUTHENTICATION_FAILURE_REPORTING_FORMAT,
        converter=DmarcReportingFormat.convert,
        validator=attr.validators.instance_of(DmarcReportingFormat),
    )
    reporting_interval = attr.ib(
        default=86400,
        converter=DmarcReportingInterval.convert,
        validator=attr.validators.instance_of(DmarcReportingInterval),
    )
    subdomain_policy = attr.ib(
        default=None,
        converter=DnsRecordTxtValueDmarcValueSubdomainPolicy.convert,
        validator=attr.validators.optional(attr.validators.instance_of(DnsRecordTxtValueDmarcValueSubdomainPolicy))
    )


class MtaStsPolicyVersion(StringEnumParsable, enum.Enum):
    STSV1 = FieldValueStringEnumParams(
        code='STSv1',
        human_readable_name='STSv1',
    )


class DnsRecordMtaStsValueVersion(FieldValueComponentParsable):
    @classmethod
    def get_canonical_name(cls):
        return 'v'

    @classmethod
    def _get_value_class(cls):
        return MtaStsPolicyVersion


class DnsRecordMtaStsValueId(FieldValueComponentString):
    @classmethod
    def get_canonical_name(cls):
        return 'id'


@attr.s
class DnsRecordTxtValueMtaSts(FieldsSemicolonSeparated):
    version = attr.ib(
        converter=DnsRecordMtaStsValueVersion.convert,
        validator=attr.validators.instance_of(DnsRecordMtaStsValueVersion)
    )
    identifier = attr.ib(
        converter=DnsRecordMtaStsValueId.convert,
        validator=attr.validators.instance_of(DnsRecordMtaStsValueId)
    )
    extensions = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(NameValuePairListSemicolonSeparated)),
        metadata={'extension': True},
    )


class TlsRptVersion(StringEnumParsable, enum.Enum):
    TLSRPTV1 = FieldValueStringEnumParams(
        code='TLSRPTv1',
        human_readable_name='TLSRPTv1',
    )


class DnsRecordTxtValueTlsRptValueVersion(FieldValueComponentParsable):
    @classmethod
    def get_canonical_name(cls):
        return 'v'

    @classmethod
    def _get_value_class(cls):
        return TlsRptVersion


class DnsRecordTxtValueTlsRptValueReportingUrlAggregated(FieldValueComponentUrl):
    @classmethod
    def get_canonical_name(cls):
        return 'rua'


@attr.s
class DnsRecordTxtValueTlsRpt(FieldsSemicolonSeparated):
    version = attr.ib(
        converter=DnsRecordTxtValueTlsRptValueVersion.convert,
        validator=attr.validators.instance_of(DnsRecordTxtValueTlsRptValueVersion)
    )
    reporting_url_aggregated = attr.ib(
        default=None,
        converter=DnsRecordTxtValueTlsRptValueReportingUrlAggregated.convert,
        validator=attr.validators.optional(
            attr.validators.instance_of(DnsRecordTxtValueTlsRptValueReportingUrlAggregated)
        ),
    )
    extensions = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(NameValuePairListSemicolonSeparated)),
        metadata={'extension': True},
    )


class SpfVersion(StringEnumParsable, enum.Enum):
    SPF1 = FieldValueStringEnumParams(
        code='spf1',
        human_readable_name='SPF1',
    )


class DnsRecordTxtValueSpfVersion(FieldValueComponentParsable):
    @classmethod
    def get_canonical_name(cls):
        return 'v'

    @classmethod
    def _get_value_class(cls):
        return SpfVersion


class SpfQualifier(StringEnumParsable, enum.Enum):
    PASS = FieldValueStringEnumParams(
        code='+',
        human_readable_name='Pass',
    )
    FAIL = FieldValueStringEnumParams(
        code='-',
        human_readable_name='Fail',
    )
    SOFTFAIL = FieldValueStringEnumParams(
        code='~',
        human_readable_name='Softfail',
    )
    NEUTRAL = FieldValueStringEnumParams(
        code='?',
        human_readable_name='Neutral',
    )


class SpfMechanism(StringEnumParsable, enum.Enum):
    ALL = FieldValueStringEnumParams(
        code='all',
        human_readable_name='All',
    )
    INCLUDE = FieldValueStringEnumParams(
        code='include',
        human_readable_name='Include',
    )
    A = FieldValueStringEnumParams(
        code='a',
        human_readable_name='A/AAAA records',
    )
    MX = FieldValueStringEnumParams(
        code='mx',
        human_readable_name='MX records',
    )
    PTR = FieldValueStringEnumParams(
        code='ptr',
        human_readable_name='PTR records',
    )
    IP4 = FieldValueStringEnumParams(
        code='ip4',
        human_readable_name='IPv4 records',
    )
    IP6 = FieldValueStringEnumParams(
        code='ip6',
        human_readable_name='IPv6 records',
    )
    EXISTS = FieldValueStringEnumParams(
        code='exists',
        human_readable_name='Exists',
    )


class SpfModifier(StringEnumParsable, enum.Enum):
    REDIRECT = FieldValueStringEnumParams(
        code='redirect',
        human_readable_name='Redirect',
    )
    EXP = FieldValueStringEnumParams(
        code='exp',
        human_readable_name='Explanation',
    )


@attr.s
class SpfDomainSpec(FieldValueSingleBase):
    @classmethod
    def _get_value_type(cls):
        return str

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_string_until_separator_or_end('value', separators=' ')

        return cls(**parser), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string(self.value)

        return composer.composed


class DnsRecordTxtValueSpfModifierKnownBase(FieldValueComponentParsable):
    @classmethod
    @abc.abstractmethod
    def get_modifier(cls):
        raise NotImplementedError()

    @classmethod
    def get_canonical_name(cls):
        return cls.get_modifier().value.code

    @classmethod
    def _get_value_class(cls):
        return SpfDomainSpec


class DnsRecordTxtValueSpfModifierRedirect(DnsRecordTxtValueSpfModifierKnownBase):
    @classmethod
    def get_modifier(cls):
        return SpfModifier.REDIRECT


class DnsRecordTxtValueSpfModifierExplanation(DnsRecordTxtValueSpfModifierKnownBase):
    @classmethod
    def get_modifier(cls):
        return SpfModifier.EXP


class DnsRecordTxtValueSpfModifierUnknown(NameValuePair):
    pass


class DnsRecordTxtValueSpfDirectiveBase(ParsableBase, Serializable):
    @classmethod
    @abc.abstractmethod
    def get_mechanism(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_qualifier_and_mechanism_name(cls, parsable):
        parser = ParserText(parsable)

        try:
            parser.parse_parsable('qualifier', SpfQualifier)
        except InvalidValue:
            pass

        mechanism = cls.get_mechanism()
        try:
            parser.parse_string('mechanism', mechanism.value.code)
        except InvalidValue as e:
            raise InvalidType from e

        return parser

    @classmethod
    def _parse_domain(cls, parser, optional, extra_separators=''):
        has_separator = True
        if optional:
            try:
                parser.parse_separator(':', min_length=1, max_length=1)
            except InvalidValue:
                has_separator = False
        else:
            parser.parse_separator(':', min_length=1, max_length=1)

        if has_separator:
            parser.parse_string_until_separator_or_end('domain', separators=' ' + extra_separators)

        return parser.get('domain', None)

    @classmethod
    def _compose_domain(cls, composer, domain):
        if domain is None:
            return

        composer.compose_separator(':')
        composer.compose_string(domain)

    @classmethod
    def _parse_ip_network(cls, parser):
        parser.parse_string('separator', ':')
        parser.parse_string_until_separator_or_end('ip_network', ' ')

        return parser['ip_network']

    @classmethod
    def _compose_ip_network(cls, composer, ip_network):
        composer.compose_separator(':')
        composer.compose_string(str(ip_network.network_address))
        if ip_network.prefixlen != ip_network.max_prefixlen:
            composer.compose_separator('/')
            composer.compose_numeric(ip_network.prefixlen)

    @classmethod
    def _parse_ip_cidr_length(cls, parser):
        has_separator = True
        try:
            parser.parse_separator('/', min_length=1, max_length=1)
        except InvalidValue:
            has_separator = False

        if has_separator:
            parser.parse_numeric('ip_cidr_length')
            ip_cidr_length = parser['ip_cidr_length']
            del parser['ip_cidr_length']
        else:
            ip_cidr_length = None

        return ip_cidr_length

    @classmethod
    def _compose_ip_cidr_length(cls, composer, ip_cidr_length):
        if ip_cidr_length is None:
            return

        composer.compose_separator('/')
        composer.compose_numeric(ip_cidr_length)

    def _compose_qualifier_and_mechanism_name(self, qualifier):
        composer = ComposerText()

        if qualifier is not None:
            composer.compose_string(qualifier.value.code)

        composer.compose_string(self.get_mechanism().value.code)

        return composer


@attr.s
class DnsRecordTxtValueSpfDirectiveAll(DnsRecordTxtValueSpfDirectiveBase):
    qualifier = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.in_(SpfQualifier))
    )

    @classmethod
    def get_mechanism(cls):
        return SpfMechanism.ALL

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_qualifier_and_mechanism_name(parsable)

        return cls(qualifier=parser.get('qualifier', None)), parser.parsed_length

    def compose(self):
        composer = self._compose_qualifier_and_mechanism_name(self.qualifier)

        return composer.composed


@attr.s
class DnsRecordTxtValueSpfDirectiveDomain(DnsRecordTxtValueSpfDirectiveBase):
    domain = attr.ib(
        converter=SpfDomainSpec.convert,
        validator=attr.validators.instance_of(SpfDomainSpec)
    )
    qualifier = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.in_(SpfQualifier))
    )

    @classmethod
    @abc.abstractmethod
    def get_mechanism(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_qualifier_and_mechanism_name(parsable)

        qualifier = parser.get('qualifier', None)
        domain = cls._parse_domain(parser, False)

        return cls(qualifier=qualifier, domain=domain), parser.parsed_length

    def compose(self):
        composer = self._compose_qualifier_and_mechanism_name(self.qualifier)

        self._compose_domain(composer, self.domain)

        return composer.composed


@attr.s
class DnsRecordTxtValueSpfDirectivePtr(DnsRecordTxtValueSpfDirectiveBase):
    domain = attr.ib(
        default=None,
        converter=SpfDomainSpec.convert,
        validator=attr.validators.optional(attr.validators.instance_of(SpfDomainSpec)),
    )
    qualifier = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.in_(SpfQualifier)),
    )

    @classmethod
    def get_mechanism(cls):
        return SpfMechanism.PTR

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_qualifier_and_mechanism_name(parsable)

        qualifier = parser.get('qualifier', None)
        domain = cls._parse_domain(parser, True)

        return cls(qualifier=qualifier, domain=domain), parser.parsed_length

    def compose(self):
        composer = self._compose_qualifier_and_mechanism_name(self.qualifier)

        self._compose_domain(composer, self.domain)

        return composer.composed


@attr.s
class DnsRecordTxtValueSpfDirectiveDomainCidr(DnsRecordTxtValueSpfDirectiveBase):
    domain = attr.ib(
        default=None,
        converter=SpfDomainSpec.convert,
        validator=attr.validators.optional(attr.validators.instance_of(SpfDomainSpec)),
    )
    ipv4_cidr_length = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(int)),
    )
    ipv6_cidr_length = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(int)),
    )
    qualifier = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.in_(SpfQualifier)),
    )

    @classmethod
    @abc.abstractmethod
    def get_mechanism(cls):
        raise NotImplementedError()

    @ipv4_cidr_length.validator
    def _validator_ipv4_cidr_length(self, attribute, value):  # pylint: disable=unused-argument
        if value is None:
            return

        if value < 0 or value > 32:
            raise InvalidValue(value, type(self), 'ipv4_cidr_length')

    @ipv6_cidr_length.validator
    def _validator_ipv6_cidr_length(self, attribute, value):  # pylint: disable=unused-argument
        if value is None:
            return

        if value < 0 or value > 128:
            raise InvalidValue(value, type(self), 'ipv6_cidr_length')

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_qualifier_and_mechanism_name(parsable)

        qualifier = parser.get('qualifier', None)
        domain = cls._parse_domain(parser, True, extra_separators='/')
        ipv4_cidr_length = cls._parse_ip_cidr_length(parser)
        ipv6_cidr_length = cls._parse_ip_cidr_length(parser)

        return cls(
            qualifier=qualifier,
            domain=domain,
            ipv4_cidr_length=ipv4_cidr_length,
            ipv6_cidr_length=ipv6_cidr_length
        ), parser.parsed_length

    def compose(self):
        composer = self._compose_qualifier_and_mechanism_name(self.qualifier)

        self._compose_domain(composer, self.domain)
        self._compose_ip_cidr_length(composer, self.ipv4_cidr_length)
        self._compose_ip_cidr_length(composer, self.ipv6_cidr_length)

        return composer.composed


class DnsRecordTxtValueSpfDirectiveInclude(DnsRecordTxtValueSpfDirectiveDomain):
    @classmethod
    def get_mechanism(cls):
        return SpfMechanism.INCLUDE


class DnsRecordTxtValueSpfDirectiveA(DnsRecordTxtValueSpfDirectiveDomainCidr):
    @classmethod
    def get_mechanism(cls):
        return SpfMechanism.A


class DnsRecordTxtValueSpfDirectiveMx(DnsRecordTxtValueSpfDirectiveDomainCidr):
    @classmethod
    def get_mechanism(cls):
        return SpfMechanism.MX


@attr.s
class DnsRecordTxtValueSpfDirectiveIp4(DnsRecordTxtValueSpfDirectiveBase):
    ipv4_network = attr.ib(
        converter=ipaddress.ip_network,
        validator=attr.validators.instance_of(ipaddress.IPv4Network)
    )
    qualifier = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.in_(SpfQualifier)),
    )

    @classmethod
    def get_mechanism(cls):
        return SpfMechanism.IP4

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_qualifier_and_mechanism_name(parsable)

        qualifier = parser.get('qualifier', None)
        ipv4_network = cls._parse_ip_network(parser)

        return cls(
            qualifier=qualifier,
            ipv4_network=ipv4_network,
        ), parser.parsed_length

    def compose(self):
        composer = self._compose_qualifier_and_mechanism_name(self.qualifier)

        self._compose_ip_network(composer, self.ipv4_network)

        return composer.composed


@attr.s
class DnsRecordTxtValueSpfDirectiveIp6(DnsRecordTxtValueSpfDirectiveBase):
    ipv6_network = attr.ib(
        converter=ipaddress.ip_network,
        validator=attr.validators.instance_of(ipaddress.IPv6Network)
    )
    qualifier = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.in_(SpfQualifier)),
    )

    @classmethod
    def get_mechanism(cls):
        return SpfMechanism.IP6

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_qualifier_and_mechanism_name(parsable)

        qualifier = parser.get('qualifier', None)
        ipv6_network = cls._parse_ip_network(parser)

        return cls(
            qualifier=qualifier,
            ipv6_network=ipv6_network,
        ), parser.parsed_length

    def compose(self):
        composer = self._compose_qualifier_and_mechanism_name(self.qualifier)

        self._compose_ip_network(composer, self.ipv6_network)

        return composer.composed


class DnsRecordTxtValueSpfDirectiveExists(DnsRecordTxtValueSpfDirectiveDomain):
    @classmethod
    def get_mechanism(cls):
        return SpfMechanism.EXISTS


class DnsRecordTxtValueSpfVariantParsable(VariantParsable):
    @classmethod
    def _get_variants(cls):
        return collections.OrderedDict([
            (SpfMechanism.ALL, [DnsRecordTxtValueSpfDirectiveAll, ]),
            (SpfMechanism.INCLUDE, [DnsRecordTxtValueSpfDirectiveInclude, ]),
            (SpfMechanism.A, [DnsRecordTxtValueSpfDirectiveA, ]),
            (SpfMechanism.MX, [DnsRecordTxtValueSpfDirectiveMx, ]),
            (SpfMechanism.PTR, [DnsRecordTxtValueSpfDirectivePtr, ]),
            (SpfMechanism.IP4, [DnsRecordTxtValueSpfDirectiveIp4, ]),
            (SpfMechanism.IP6, [DnsRecordTxtValueSpfDirectiveIp6, ]),
            (SpfMechanism.EXISTS, [DnsRecordTxtValueSpfDirectiveExists, ]),
            (SpfModifier.REDIRECT, [DnsRecordTxtValueSpfModifierRedirect, ]),
            (SpfModifier.EXP, [DnsRecordTxtValueSpfModifierExplanation, ]),
        ])


@attr.s
class DnsRecordTxtValueSpf(ParsableBase, Serializable):
    terms = attr.ib(
        validator=attr.validators.deep_iterable(member_validator=attr.validators.instance_of((
            DnsRecordTxtValueSpfDirectiveBase,
            DnsRecordTxtValueSpfModifierKnownBase,
            DnsRecordTxtValueSpfModifierUnknown,
        )))
    )
    version = attr.ib(
        default=DnsRecordTxtValueSpfVersion(SpfVersion.SPF1),
        converter=DnsRecordTxtValueSpfVersion.convert,
        validator=attr.validators.instance_of(DnsRecordTxtValueSpfVersion)
    )

    def _asdict(self):
        terms = []
        for term in self.terms:
            if isinstance(term, DnsRecordTxtValueSpfModifierKnownBase):
                terms.append((term.get_modifier(), term.value.value))
            elif isinstance(term, DnsRecordTxtValueSpfDirectiveBase):
                terms.append((term.get_mechanism(), term._asdict()))
            elif isinstance(term, DnsRecordTxtValueSpfModifierUnknown):
                terms.append((term.name, term.value))
            else:
                raise NotImplementedError()

        return collections.OrderedDict([
            ('Version', self.version),
            ('Terms', collections.OrderedDict(terms)),
        ])

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        try:
            parser.parse_parsable('version', DnsRecordTxtValueSpfVersion)
        except InvalidValue as e:
            raise InvalidType from e

        terms = []
        while parser.unparsed_length:
            parser.parse_separator(' ')

            try:
                parser.parse_parsable('term', DnsRecordTxtValueSpfVariantParsable)
                term = parser['term']
            except InvalidValue:
                parser.parse_string_until_separator_or_end('term', ' ')
                term_parser = ParserText(parser['term'].encode('ascii'))
                term_parser.parse_parsable('value', DnsRecordTxtValueSpfModifierUnknown)
                term = term_parser['value']

            terms.append(term)
            del parser['term']

        return cls(version=parser['version'], terms=terms), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_parsable(self.version)

        for term in self.terms:
            composer.compose_separator(' ')
            composer.compose_parsable(term)

        return composer.composed



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines

import abc
import collections.abc
import datetime
import enum
import struct
import time
import typing

import attr

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.types import CryptoDataEnumBase, CryptoDataEnumCodedBase

from cryptoparser.common.exception import InvalidType, NotEnoughData, TooMuchData
import cryptoparser.common.utils


class ParsableBaseNoABC():
    @classmethod
    def parse_mutable(cls, parsable):
        parsed_object, parsed_length = cls._parse(parsable)
        del parsable[:parsed_length]
        return parsed_object

    @classmethod
    def parse_immutable(cls, parsable):
        parsed_object, parsed_length = cls._parse(parsable)
        return parsed_object, parsed_length

    @classmethod
    def parse_exact_size(cls, parsable):
        parsed_object, parsed_length = cls._parse(parsable)
        if len(parsable) > parsed_length:
            raise TooMuchData(parsed_length)

        return parsed_object

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class ParsableBase(ParsableBaseNoABC, metaclass=abc.ABCMeta):
    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class ParserCRLF(ParsableBase):
    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_string('crlf', '\r\n')

        return ParserCRLF(), 2

    def compose(self):
        return b'\r\n'


_SIZE_TO_FORMAT = {
    1: 'B',
    2: 'H',
    3: 'I',
    4: 'I',
    8: 'Q',
}


class ByteOrder(enum.Enum):
    NATIVE = '='
    LITTLE_ENDIAN = '<'
    BIG_ENDIAN = '>'
    NETWORK = '!'


@attr.s
class ParserBase(collections.abc.Mapping):
    _parsable: typing.Union[bytes, bytearray] = attr.ib(
        converter=bytes,
        validator=attr.validators.instance_of((bytes, bytearray))
    )
    _parsed_length: int = attr.ib(init=False, default=0)
    _parsed_values: typing.Dict[str, typing.Any] = attr.ib(init=False, default=None)

    def __attrs_post_init__(self):
        if self._parsed_values is None:
            self._parsed_values = {}

    def __len__(self):
        return len(self._parsed_values)

    def __iter__(self):
        return iter(self._parsed_values)

    def __getitem__(self, key):
        return self._parsed_values[key]

    def __delitem__(self, key):
        del self._parsed_values[key]

    @property
    def parsed_length(self):
        return self._parsed_length

    @property
    def unparsed(self):
        return self._parsable[self._parsed_length:]

    @property
    def unparsed_length(self):
        return len(self._parsable) - self._parsed_length

    @abc.abstractmethod
    def _parse_numeric(self, name, converter, item_size):
        raise NotImplementedError()

    def parse_parsable(self, name, parsable_class, item_size=None):
        if item_size is None:
            parsed_object, parsed_length = parsable_class.parse_immutable(
                self._parsable[self._parsed_length:]
            )
        else:
            parsable_length, _ = self._parse_numeric(name, int, item_size)
            parsable_length = parsable_length[0]
            parsed_object = parsable_class.parse_exact_size(
                self._parsable[self._parsed_length + item_size:self._parsed_length + parsable_length + item_size]
            )
            parsed_length = item_size + parsable_length

        self._parsed_length += parsed_length
        self._parsed_values[name] = parsed_object

    def _parse_string_by_length(
            self,
            name,
            item_min_length,
            item_max_length,
            encoding,
            converter
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        if item_min_length > self.unparsed_length:
            raise NotEnoughData(item_min_length - self.unparsed_length)

        if item_max_length is None:
            parsable_length = len(self._parsable) - self.parsed_length
        else:
            parsable_length = min(item_max_length, self.unparsed_length)

        value = self._parsable[self._parsed_length:self._parsed_length + parsable_length]
        try:
            value = value.decode(encoding)
            if converter != str:
                value = converter(value)
            self._parsed_values[name] = value
        except UnicodeError as e:
            raise InvalidValue(value, converter, name) from e
        except ValueError as e:
            raise InvalidValue(value, converter, name) from e

        return value, parsable_length


class ParserText(ParserBase):
    def __init__(self, parsable, encoding='ascii'):
        super().__init__(parsable)
        self._encoding = encoding

    def _check_separators(  # pylint: disable=too-many-arguments,too-many-positional-arguments
            self,
            name,
            count_offset,
            separators,
            min_count,
            max_count
    ):
        separators = separators.encode(self._encoding)

        count = 0
        actual_offset = count_offset
        while actual_offset < len(self._parsable) and self._parsable[actual_offset:actual_offset + 1] in separators:
            actual_offset += 1
            count += 1

            if max_count is not None and count > max_count:
                raise InvalidValue(self._parsable[count_offset:], type(self), name)

        if min_count is not None and count < min_count:
            raise InvalidValue(self._parsable[count_offset:], type(self), name)

        return actual_offset - count_offset

    def parse_separator(self, separator, min_length=1, max_length=None):
        self._parsed_length += self._check_separators(
            'separator', self._parsed_length, separator, min_length, max_length
        )

    def _parse_numeric_array(  # pylint: disable=too-many-arguments,too-many-positional-arguments
            self, name, item_num, separator, converter, is_floating):
        value = []
        floating_point_found = False
        last_item_offset = self._parsed_length
        item_offset = self._parsed_length
        while True:
            while item_offset < len(self._parsable) and self._parsable[item_offset:item_offset + 1].isdigit():
                item_offset += 1

            if item_offset == last_item_offset:
                raise InvalidValue(self._parsable[self._parsed_length:], type(self), name)

            if (is_floating and
                    not floating_point_found and item_offset < len(self._parsable) and
                    self._parsable[item_offset] == ord('.')):
                item_offset += 1
                floating_point_found = True
                continue

            value.append(converter(self._parsable[last_item_offset:item_offset]))

            if item_offset == len(self._parsable) or (item_num is not None and len(value) == item_num):
                break

            if separator:
                try:
                    item_offset += self._check_separators(name, item_offset, separator, 1, 1)
                except InvalidValue as e:
                    raise InvalidValue(self._parsable[self._parsed_length:item_offset], type(self), name) from e

            last_item_offset = item_offset
            floating_point_found = False

        return value, item_offset - self._parsed_length

    def _parse_numeric(self, name, converter, item_size):
        raise NotImplementedError()

    def parse_numeric(self, name, converter=int):
        value, parsed_length = self._parse_numeric_array(name, 1, None, converter, False)
        self._parsed_values[name] = value[0]
        self._parsed_length += parsed_length

    def parse_float(self, name, converter=float):
        value, parsed_length = self._parse_numeric_array(name, 1, None, converter, True)
        self._parsed_values[name] = value[0]
        self._parsed_length += parsed_length

    def parse_numeric_array(self, name, item_num, separator, converter=int):
        value, parsed_length = self._parse_numeric_array(name, item_num, separator, converter, False)

        self._parsed_values[name] = value
        self._parsed_length += parsed_length

    def parse_bool(self, name):
        for string_value, bool_value in (('yes', True), ('no', False)):
            try:
                self.parse_string(name, string_value)
            except InvalidValue:
                pass
            else:
                self._parsed_values[name] = bool_value
                break
        else:
            raise InvalidValue(self._parsable[self._parsed_length:], type(self), name)

    def parse_string(self, name, value):
        min_length = len(value)
        max_length = min_length
        try:
            actual_value, parsed_length = self._parse_string_by_length(
                name, min_length, max_length, self._encoding, str
            )
        except NotEnoughData as e:
            assert e.bytes_needed is not None
            raise InvalidValue(self._parsable[self._parsed_length:min_length - e.bytes_needed], type(self), name) from e

        if value != actual_value:
            raise InvalidValue(self._parsable[self._parsed_length:self._parsed_length + max_length], type(self), name)

        self._parsed_values[name] = value
        self._parsed_length += parsed_length

    def parse_string_by_length(self, name, min_length=1, max_length=None, item_class=str):
        value, parsed_length = self._parse_string_by_length(name, min_length, max_length, self._encoding, item_class)
        self._parsed_values[name] = value
        self._parsed_length += parsed_length

    def _apply_item_class(  # pylint: disable=too-many-arguments,too-many-positional-arguments
            self,
            name,
            item_offset,
            item_end,
            separator,
            item_class,
            fallback_class,
            may_end):
        try:
            if not isinstance(item_class, type):
                item = item_class(self._parsable[item_offset:item_end].decode(self._encoding))
            elif issubclass(item_class, CryptoDataEnumCodedBase):
                item = item_class.from_code(self._parsable[item_offset:item_end].decode(self._encoding))
            elif issubclass(item_class, ParsableBaseNoABC):
                item, parsed_length = item_class.parse_immutable(self._parsable[item_offset:item_end])
                item_end = item_offset + parsed_length
            elif issubclass(item_class, str):
                item = self._parsable[item_offset:item_end].decode(self._encoding)
            else:
                item = item_class(self._parsable[item_offset:item_end].decode(self._encoding))
        except (InvalidValue, ValueError, UnicodeError) as e:
            if fallback_class is not None:
                parsed_value, parsed_length = self._parse_string_until_separator(
                    name, item_offset, separator, fallback_class, None, may_end
                )
                item_offset += parsed_length
                return parsed_value

            raise InvalidValue(self._parsable[item_offset:], type(self), name) from e

        return item

    def _parse_string_until_separator(  # pylint: disable=too-many-arguments,too-many-positional-arguments
            self,
            name,
            item_offset,
            separators,
            item_class,
            fallback_class,
            may_end=False,
            separator_spaces=''
    ):
        item_end = None
        byte_separators = [separator.encode(self._encoding) for separator in separators]

        for separator_end in range(item_offset, len(self._parsable) + 1):
            for separator in byte_separators:
                if self._parsable[item_offset:separator_end].endswith(separator):
                    item_end = separator_end - len(separator)
                    break
            if item_end is not None:
                break
        else:
            if not may_end:
                raise InvalidValue(self._parsable[item_offset:], type(item_class), name)

            item_end = len(self._parsable)

        separator_space_count = 0
        byte_separator_spaces = separator_spaces.encode(self._encoding)
        while (item_end > item_offset and
                self._parsable[
                    item_end - separator_space_count - 1:
                    item_end - separator_space_count
                ] in byte_separator_spaces):
            separator_space_count += 1

        item = self._apply_item_class(
            name, item_offset, item_end - separator_space_count, separators, item_class, fallback_class, may_end
        )

        return item, item_end - item_offset - separator_space_count

    def parse_string_until_separator(self, name, separators, item_class=str, fallback_class=None):
        parsed_value, parsed_length = self._parse_string_until_separator(
            name, self._parsed_length, separators, item_class, fallback_class, False
        )

        self._parsed_values[name] = parsed_value
        self._parsed_length += parsed_length

    def parse_string_until_separator_or_end(self, name, separators, item_class=str, fallback_class=None):
        parsed_value, parsed_length = self._parse_string_until_separator(
            name, self._parsed_length, separators, item_class, fallback_class, True
        )

        self._parsed_values[name] = parsed_value
        self._parsed_length += parsed_length

    def _parse_string_array(
            self,
            name,
            separator,
            max_item_num=None,
            item_class=str,
            fallback_class=None,
            separator_spaces='',
            skip_empty=False
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        value = []
        item_offset = self._parsed_length
        max_separator_count = None if skip_empty else 1

        if separator_spaces:
            item_offset += self._check_separators('separator', item_offset, separator_spaces, None, None)

        while True:
            parsed_value, parsed_length = self._parse_string_until_separator(
                name, item_offset, separator, str, None, True, separator_spaces
            )
            if parsed_length:
                if isinstance(item_class, type) and issubclass(item_class, ParsableBase):
                    parsed_value = item_class.parse_exact_size(parsed_value.encode(self._encoding))
                else:
                    parsed_value = self._apply_item_class(
                        name,
                        item_offset,
                        item_offset + parsed_length,
                        separator,
                        item_class,
                        fallback_class,
                        True
                    )
                value.append(parsed_value)
                item_offset += parsed_length
            elif not skip_empty:
                raise InvalidValue(self._parsable[item_offset:], type(self), name)

            if separator_spaces:
                item_offset += self._check_separators('separator', item_offset, separator_spaces, None, None)

            if item_offset == len(self._parsable):
                break

            item_offset += self._check_separators(name, item_offset, separator, 1, max_separator_count)
            if separator_spaces:
                item_offset += self._check_separators('separator', item_offset, separator_spaces, None, None)

            if item_offset == len(self._parsable):
                break
            if max_item_num is not None and len(value) == max_item_num:
                break

        self._parsed_values[name] = value
        self._parsed_length = item_offset

    def parse_string_array(
            self,
            name,
            separator,
            item_class=str,
            fallback_class=None,
            separator_spaces='',
            skip_empty=False,
            max_item_num=None,
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self._parse_string_array(
            name,
            separator,
            max_item_num=max_item_num,
            item_class=item_class,
            fallback_class=fallback_class,
            separator_spaces=separator_spaces,
            skip_empty=skip_empty
        )

    def parse_date_time(self, name):
        try:
            value = self._parsable[self._parsed_length:]
            value_str = value.decode(self._encoding)

            date_time = datetime.datetime.strptime(value_str, '%a, %d %b %Y %H:%M:%S %z')
        except ValueError:
            try:
                date_time = datetime.datetime.strptime(value_str, '%a, %d %b %Y %H:%M:%S GMT')
                date_time = date_time.replace(tzinfo=datetime.timezone.utc)
            except ValueError as e:
                raise InvalidValue(value, type(self), 'value') from e

        self._parsed_values[name] = date_time
        self._parsed_length = len(self._parsable)

    def parse_time_delta(self, name):
        value, parsed_length = self._parse_numeric_array(name, 1, None, int, False)

        try:
            time_delta = datetime.timedelta(seconds=value[0])
        except OverflowError as e:
            raise InvalidValue(value[0], type(self), 'value') from e

        self._parsed_values[name] = time_delta
        self._parsed_length += parsed_length


@attr.s
class ParserBinary(ParserBase):
    byte_order: ByteOrder = attr.ib(default=ByteOrder.NETWORK, validator=attr.validators.in_(ByteOrder))

    def parse_timestamp(self, name, milliseconds=False, item_size=8):
        value, parsed_length = self._parse_numeric_array(name, 1, item_size, int)

        self._parsed_length += parsed_length
        if value[0] == (2 ** (8 * item_size) - 1):
            self._parsed_values[name] = None
        else:
            value = value[0]
            if milliseconds:
                millis = value % 1000
                value //= 1000
            value = datetime.datetime.fromtimestamp(0x00000000ffffffff & value, datetime.timezone.utc)
            if milliseconds:
                value += datetime.timedelta(milliseconds=millis)
            self._parsed_values[name] = value

    def _parse_numeric_array(self, name, item_num, item_size, item_numeric_class):
        if self._parsed_length + (item_num * item_size) > len(self._parsable):
            raise NotEnoughData(bytes_needed=(item_num * item_size) - self.unparsed_length)

        if item_size in _SIZE_TO_FORMAT:
            value = []
            for item_offset in range(self._parsed_length, self._parsed_length + (item_num * item_size), item_size):
                item_bytes = self._parsable[item_offset:item_offset + item_size]
                if item_size == 3:
                    if self.byte_order in [ByteOrder.BIG_ENDIAN, ByteOrder.NETWORK]:
                        item_bytes = b'\x00' + item_bytes
                    else:
                        item_bytes = item_bytes + b'\x00'

                item = struct.unpack(
                    self.byte_order.value + _SIZE_TO_FORMAT[item_size],
                    item_bytes
                )[0]
                try:
                    value.append(item_numeric_class(item))
                except ValueError as e:
                    raise InvalidValue(item, item_numeric_class, name) from e
        else:
            raise NotImplementedError()

        return value, item_num * item_size

    def _parse_numeric(self, name, converter, item_size):
        return self._parse_numeric_array(name, 1, item_size, converter)

    def parse_numeric(self, name, size, converter=int):
        value, parsed_length = self._parse_numeric_array(name, 1, size, converter)

        self._parsed_length += parsed_length
        self._parsed_values[name] = value[0]

    def parse_numeric_enum_coded(self, name, enum_class):
        param = list(enum_class)[0].value
        value, parsed_length = self._parse_numeric_array(name, 1, param.get_code_size(), enum_class.from_code)

        self._parsed_length += parsed_length
        self._parsed_values[name] = value[0]

    def parse_numeric_array(self, name, item_num, item_size, converter=int):
        value, parsed_length = self._parse_numeric_array(name, item_num, item_size, converter)

        self._parsed_length += parsed_length
        self._parsed_values[name] = value

    def parse_numeric_flags(self, name, size, flags_class, shift_left=0):
        value, parsed_length = self._parse_numeric_array(name, 1, size, int)
        value = {
            flags_class(flag & (value[0] << shift_left))
            for flag in flags_class
            if flag & (value[0] << shift_left)
        }

        self._parsed_length += parsed_length
        self._parsed_values[name] = value

    def _parse_mpint(self, mpint_length, mpint_offset, negative):
        if mpint_length % 4:
            pad_byte = bytes((0xff,)) if negative else bytes((0x00,))
            pad_bytes = (4 - (mpint_length % 4)) * pad_byte
        else:
            pad_bytes = b''

        parsable = pad_bytes + self._parsable[self._parsed_length + mpint_offset:]
        parser = ParserBinary(parsable)
        parser.parse_numeric_array('mpint_part_array', (mpint_length + len(pad_bytes)) // 4, 4, int)

        value = 0
        for mpint_part in parser['mpint_part_array']:
            value = (value << 32) + mpint_part
        if negative:
            complement = 1 << (8 * (mpint_length + len(pad_bytes)))
            value -= complement

        return value

    def parse_mpint(self, name, mpint_length):
        value = self._parse_mpint(mpint_length, 0, False)

        self._parsed_values[name] = value
        self._parsed_length += mpint_length

    def parse_ssh_mpint(self, name):
        if self.unparsed_length < 4:
            raise NotEnoughData(bytes_needed=4 - self.unparsed_length)

        mpint_length, parsed_length = self._parse_numeric_array(name, 1, 4, int)
        mpint_length = mpint_length[0]

        negative = (mpint_length and self._parsable[self._parsed_length + 4] >= 0x80)

        value = self._parse_mpint(mpint_length, 4, negative)

        self._parsed_values[name] = value
        self._parsed_length += parsed_length + mpint_length

    def _parse_bytes(self, size):
        if self.unparsed_length < size:
            raise NotEnoughData(bytes_needed=size - self.unparsed_length)

        return self._parsable[self._parsed_length: self._parsed_length + size]

    def parse_bytes(self, name, size, converter=bytearray):
        value, parsed_length = self._parse_numeric_array(name, 1, size, int)
        value = value[0]

        self._parsed_length += parsed_length
        try:
            parsed_bytes = self._parse_bytes(value)
        except NotEnoughData:
            self._parsed_length -= parsed_length
            raise

        try:
            self._parsed_values[name] = converter(parsed_bytes)
        except ValueError as e:
            raise InvalidValue(value, converter, name) from e
        self._parsed_length += len(parsed_bytes)

    def parse_raw(self, name, size, converter=bytearray):
        parsed_bytes = self._parse_bytes(size)

        try:
            self._parsed_values[name] = converter(parsed_bytes)
        except ValueError as e:
            raise InvalidValue(parsed_bytes, converter, name) from e

        self._parsed_length += size

    def parse_string(self, name, item_size, encoding, converter=str):
        value, parsed_length = self._parse_numeric_array(name, 1, item_size, int)
        value = value[0]
        self._parsed_length += parsed_length

        try:
            value, parsed_length = self._parse_string_by_length(name, value, value, encoding, converter)
            self._parsed_length += parsed_length
            self._parsed_values[name] = value
        except InvalidValue as e:
            raise e

    def parse_string_null_terminated(self, name, encoding, converter=str):
        try:
            length = next(iter([
                i
                for i, value in enumerate(self._parsable[self._parsed_length:])
                if value == 0
            ]))
        except StopIteration as e:
            raise InvalidValue(self._parsable[self._parsed_length:], str, name) from e

        value, parsed_length = self._parse_string_by_length(name, length, length, encoding, converter)
        self._parsed_length += parsed_length + 1
        self._parsed_values[name] = value

    def _parse_parsable_derived_array(self, items_size, item_classes, fallback_class=None):
        if items_size > self.unparsed_length:
            raise NotEnoughData(bytes_needed=items_size - self.unparsed_length)

        items = []
        unparsed_bytes = self._parsable[self._parsed_length:self._parsed_length + items_size]

        while unparsed_bytes:
            for item_class in item_classes:
                try:
                    item, parsed_length = item_class.parse_immutable(unparsed_bytes)
                    break
                except InvalidValue:
                    pass
            else:
                if fallback_class is not None:
                    item, parsed_length = fallback_class.parse_immutable(unparsed_bytes)
                else:
                    raise ValueError(unparsed_bytes)

            unparsed_bytes = unparsed_bytes[parsed_length:]
            items.append(item)

        return items, items_size

    def parse_parsable_array(self, name, items_size, item_class, fallback_class=None):
        try:
            items, items_size = self._parse_parsable_derived_array(items_size, [item_class, ], fallback_class)
        except NotEnoughData as e:
            raise e
        except ValueError as e:
            raise InvalidValue(e.args[0], item_class, name) from e

        self._parsed_values[name] = items
        self._parsed_length += items_size

    def parse_parsable_derived_array(self, name, items_size, item_base_class, fallback_class=None):
        item_classes = cryptoparser.common.utils.get_leaf_classes(item_base_class)
        try:
            items, items_size = self._parse_parsable_derived_array(items_size, item_classes, fallback_class)
        except NotEnoughData as e:
            raise e
        except ValueError as e:
            raise InvalidValue(e.args[0], item_base_class, name) from e

        self._parsed_values[name] = items
        self._parsed_length += items_size

    def _remove_trailing_separator(self, items, name):
        if not isinstance(items[-1], ParserCRLF):
            raise InvalidValue(self.unparsed, type(self), name)
        del items[-1]

    def parse_parsable_list(self, name, item_class, fallback_class=None, separator_class=ParserCRLF):
        try:
            items, items_size = self._parse_parsable_derived_array(
                self.unparsed_length, [separator_class, item_class], fallback_class
            )
        except ValueError as e:
            raise InvalidValue(e.args[0], item_class, name) from e

        if not items:
            raise NotEnoughData(2)
        self._remove_trailing_separator(items, name)

        real_items = []
        while items:
            if len(items) % 2 == 0:
                self._remove_trailing_separator(items, name)
            else:
                if isinstance(items[-1], separator_class):
                    raise InvalidValue(self.unparsed, type(self), name)

                real_items.insert(0, items[-1])
                del items[-1]

        self._parsed_values[name] = real_items
        self._parsed_length += items_size

    def parse_variant(self, name, variant):
        parsed_object, value_length = variant.parse(self._parsable[self._parsed_length:])

        self._parsed_values[name] = parsed_object
        self._parsed_length += value_length


@attr.s
class ComposerBase():
    _composed = attr.ib(init=False, default=bytes())

    @property
    def composed(self):
        return self._composed

    @property
    def composed_length(self):
        return len(self._composed)

    def _compose_string_array(self, values, encoding, separator):
        separator = bytearray(separator.encode(encoding))
        composed_str = bytearray()

        for value in values:
            try:
                if isinstance(value, ParsableBaseNoABC):
                    composed_str += value.compose()
                else:
                    composed_str += str(value).encode(encoding)
            except UnicodeError as e:
                raise InvalidValue(value, type(self)) from e

            composed_str += separator

        self._composed += composed_str[:len(composed_str) - len(separator)]


class ComposerText(ComposerBase):
    def __init__(self, encoding='ascii'):
        super().__init__()
        self._encoding = encoding

    def _compose_numeric_array(self, values, separator):
        composed_str = str()

        for value in values:
            composed_str += f'{value:d}{separator}'

        self._composed += composed_str[:len(composed_str) - len(separator)].encode(self._encoding)

    def compose_numeric(self, value):
        self._compose_numeric_array([value, ], separator='')

    def compose_numeric_array(self, values, separator):
        self._compose_numeric_array(values, separator)

    def compose_bool(self, value):
        self.compose_string('yes' if value else 'no')

    def compose_string(self, value):
        self._compose_string_array([value, ], encoding=self._encoding, separator='')

    def compose_string_array(self, value, separator=','):
        self._compose_string_array(value, encoding=self._encoding, separator=separator)

    @staticmethod
    def _compose_parsable(value):
        return value.compose()

    def compose_parsable(self, value):
        self._composed += self._compose_parsable(value)

    def compose_parsable_array(self, values, separator=',', fallback_class=None):
        separator = separator.encode(self._encoding)

        composed_items = []
        for item in values:
            if isinstance(item, (ComposerBase, ParsableBase, ParsableBaseNoABC)):
                composed_item = self._compose_parsable(item)
            elif isinstance(item, CryptoDataEnumBase):
                composed_item = item.value.code.encode(self._encoding)
            elif fallback_class is not None and isinstance(item, fallback_class):
                composed_item = item.encode(self._encoding)
            else:
                raise InvalidType()

            composed_items.append(composed_item)

        self._composed += bytearray(separator).join(composed_items)

    def compose_separator(self, value):
        self.compose_string(value)

    def compose_date_time(self, value, fmt):
        self.compose_string(value.strftime(fmt))

    def compose_time_delta(self, value):
        self.compose_numeric(int(value.total_seconds()))


@attr.s
class ComposerBinary(ComposerBase):
    byte_order: ByteOrder = attr.ib(default=ByteOrder.NETWORK, validator=attr.validators.in_(ByteOrder))

    def compose_timestamp(self, value, milliseconds=False, item_size=8):
        if value is None:
            timestamp = 0xffffffffffffffff
        else:
            timestamp = int(time.mktime(value.timetuple())) - time.timezone

            if milliseconds:
                timestamp *= 1000
                timestamp += value.microsecond // 1000

        return self._compose_numeric_array([timestamp, ], item_size)

    def _compose_numeric_array(self, values, item_size):
        composed_bytes = bytearray()

        for value in values:
            try:
                packed_bytes = struct.pack(
                    self.byte_order.value + _SIZE_TO_FORMAT[item_size],
                    value
                )

                if item_size == 3:
                    if self.byte_order in [ByteOrder.BIG_ENDIAN, ByteOrder.NETWORK]:
                        composed_bytes += packed_bytes[1:]
                    else:
                        composed_bytes += packed_bytes[:3]
                else:
                    composed_bytes += packed_bytes
            except struct.error as e:
                raise InvalidValue(value, int) from e

        self._composed += composed_bytes

    def compose_numeric(self, value, size):
        self._compose_numeric_array([value, ], size)

    def compose_numeric_enum_coded(self, value):
        self.compose_numeric_array_enum_coded([value, ])

    def compose_numeric_array(self, values, item_size):
        self._compose_numeric_array(values, item_size)

    def compose_numeric_array_enum_coded(self, values):
        if not values:
            return

        self._compose_numeric_array(map(lambda value: value.value.code, values), values[0].value.get_code_size())

    def compose_numeric_flags(self, values, item_size, shift_right=0):
        flag = 0
        for value in values:
            flag |= value >> shift_right
        self._compose_numeric_array([flag, ], item_size)

    @staticmethod
    def _compose_mpint(value, length, byte_order):
        negative = value < 0

        if negative:
            positive_value = (1 << ((value.bit_length() // 8 * 8) + 8)) + value
        else:
            positive_value = value

        value_numeric_array = []
        for mpint_offset in range(0, length * 32, 32):
            value_numeric_array.append(positive_value >> mpint_offset & 0xffffffff)

        composer = ComposerBinary(byte_order=byte_order)
        composer.compose_numeric_array(reversed(value_numeric_array), 4)

        mpint_bytes = composer.composed_bytes.lstrip(b'\x00')
        if byte_order in [ByteOrder.LITTLE_ENDIAN, ByteOrder.NATIVE]:
            mpint_bytes = mpint_bytes.rstrip(b'\x00')

        return mpint_bytes

    def compose_mpint(self, value, length):
        mpint_bytes = self._compose_mpint(value, length, self.byte_order)
        if length < len(mpint_bytes):
            raise InvalidValue(length, type(self), 'mpint_length')

        pad_byte = b'\xff' if value < 0 else b'\x00'
        if self.byte_order in [ByteOrder.BIG_ENDIAN, ByteOrder.NETWORK]:
            self.compose_raw((length - len(mpint_bytes)) * pad_byte)
            self.compose_raw(mpint_bytes)
        else:
            self.compose_raw(mpint_bytes)
            self.compose_raw((length - len(mpint_bytes)) * pad_byte)

    def compose_ssh_mpint(self, value):
        negative = value < 0
        length = value.bit_length() // 32
        if value.bit_length() % 32:
            length += 1

        mpint_bytes = self._compose_mpint(value, length, self.byte_order)

        if mpint_bytes and bool(mpint_bytes[0] & 0x80) != negative:
            pad_byte = b'\xff' if negative else b'\x00'
        else:
            pad_byte = b''

        self.compose_numeric(len(pad_byte) + len(mpint_bytes), 4)
        self.compose_raw(pad_byte)
        self.compose_raw(mpint_bytes)

    def compose_parsable(self, value, item_size=None):
        composed = value.compose()
        if item_size is not None:
            self.compose_numeric(len(composed), item_size)
        self._composed += composed

    def compose_parsable_array(self, values, separator=bytearray()):
        self._composed += separator.join(map(lambda item: item.compose(), values))

    def compose_bytes(self, value, item_size, converter=bytearray):
        value_bytes = converter(value)
        self._compose_numeric_array([len(value_bytes), ], item_size)
        self.compose_raw(value_bytes)

    def compose_raw(self, value):
        self._composed += value

    def compose_string(self, value, encoding, item_size):
        try:
            value = value.encode(encoding)
        except UnicodeError as e:
            raise InvalidValue(value, type(self)) from e

        self.compose_bytes(value, item_size)

    def compose_string_null_terminated(self, value, encoding):
        try:
            value = value.encode(encoding)
        except UnicodeError as e:
            raise InvalidValue(value, type(self)) from e

        self.compose_raw(value)
        self.compose_raw(b'\x00')

    def compose_string_enum_coded(self, value, item_size):
        self.compose_string(value.value.code, 'ascii', item_size)

    @property
    def composed_bytes(self):
        return bytearray(self._composed)

    @property
    def composed_length(self):
        return len(self._composed)



# SPDX-License-Identifier: MPL-2.0
# pylint: disable=too-many-lines
# -*- coding: utf-8 -*-

import os

import unittest

import datetime

from cryptodatahub.common.algorithm import Hash
from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.exception import InvalidType
from cryptoparser.httpx.header import (
    ContentSecurityPolicyDirectiveBaseUri,
    ContentSecurityPolicyDirectiveBlockAllMixedContent,
    ContentSecurityPolicyDirectiveChildSrc,
    ContentSecurityPolicyDirectiveConnectSrc,
    ContentSecurityPolicyDirectiveDefaultSrc,
    ContentSecurityPolicyDirectiveFontSrc,
    ContentSecurityPolicyDirectiveFormAction,
    ContentSecurityPolicyDirectiveFrameAncestors,
    ContentSecurityPolicyDirectiveFrameSrc,
    ContentSecurityPolicyDirectiveImgSrc,
    ContentSecurityPolicyDirectiveManifestSrc,
    ContentSecurityPolicyDirectiveMediaSrc,
    ContentSecurityPolicyDirectiveObjectSrc,
    ContentSecurityPolicyDirectivePluginTypes,
    ContentSecurityPolicyDirectivePrefetchSrc,
    ContentSecurityPolicyDirectiveReferrer,
    ContentSecurityPolicyDirectiveReportTo,
    ContentSecurityPolicyDirectiveReportUri,
    ContentSecurityPolicyDirectiveRequireTrustedTypesFor,
    ContentSecurityPolicyDirectiveSandbox,
    ContentSecurityPolicyDirectiveScriptSrc,
    ContentSecurityPolicyDirectiveScriptSrcAttr,
    ContentSecurityPolicyDirectiveScriptSrcElem,
    ContentSecurityPolicyDirectiveStyleSrc,
    ContentSecurityPolicyDirectiveStyleSrcAttr,
    ContentSecurityPolicyDirectiveStyleSrcElem,
    ContentSecurityPolicyDirectiveUpgradeInsecureRequests,
    ContentSecurityPolicyDirectiveWebrtc,
    ContentSecurityPolicyDirectiveWorkerSrc,
    ContentSecurityPolicyReferrerPolicy,
    ContentSecurityPolicySourceHash,
    ContentSecurityPolicySourceHost,
    ContentSecurityPolicySourceKeyword,
    ContentSecurityPolicySourceNonce,
    ContentSecurityPolicySourceScheme,
    ContentSecurityPolicyTrustedTypeSinkGroup,
    ContentSecurityPolicyWebRtcType,
    FieldValueMimeType,
    HttpHeaderFields,
    HttpHeaderFieldAge,
    HttpHeaderFieldCacheControlResponse,
    HttpHeaderFieldContentType,
    HttpHeaderFieldContentSecurityPolicy,
    HttpHeaderFieldContentSecurityPolicyReportOnly,
    HttpHeaderFieldDate,
    HttpHeaderFieldETag,
    HttpHeaderFieldExpectCT,
    HttpHeaderFieldExpectStaple,
    HttpHeaderFieldExpires,
    HttpHeaderFieldLastModified,
    HttpHeaderFieldName,
    HttpHeaderFieldNetworkErrorLogging,
    HttpHeaderFieldPragma,
    HttpHeaderFieldPublicKeyPinning,
    HttpHeaderFieldReferrerPolicy,
    HttpHeaderFieldServer,
    HttpHeaderFieldSetCookie,
    HttpHeaderFieldSTS,
    HttpHeaderFieldUnparsed,
    HttpHeaderFieldValueCacheControlResponse,
    HttpHeaderFieldValueContentSecurityPolicy,
    HttpHeaderFieldValueContentType,
    HttpHeaderFieldValueExpectCT,
    HttpHeaderFieldValueExpectStaple,
    HttpHeaderFieldValueNetworkErrorLogging,
    HttpHeaderFieldValuePragma,
    HttpHeaderFieldValuePublicKeyPinning,
    HttpHeaderFieldValueReferrerPolicy,
    HttpHeaderFieldValueSetCookie,
    HttpHeaderFieldValueSTS,
    HttpHeaderFieldValueXContentTypeOptions,
    HttpHeaderFieldValueXFrameOptions,
    HttpHeaderFieldValueXXSSProtection,
    HttpHeaderFieldXContentSecurityPolicy,
    HttpHeaderFieldXContentTypeOptions,
    HttpHeaderFieldXFrameOptions,
    HttpHeaderFieldXXSSProtection,
    HttpHeaderReferrerPolicy,
    HttpHeaderPragma,
    HttpHeaderSetCookieComponentSameSite,
    HttpHeaderXContentTypeOptions,
    HttpHeaderXFrameOptions,
    HttpHeaderXXSSProtectionMode,
    HttpHeaderXXSSProtectionState,
    MimeTypeRegistry,
)

from .classes import TestCasesBasesHttpHeader


class TestHttpHeaderFieldValueCacheControlResponse(
        TestCasesBasesHttpHeader.MinimalHeader,
        TestCasesBasesHttpHeader.FullHeader,
        TestCasesBasesHttpHeader.CaseInsensitiveHeader):

    _header_minimal = HttpHeaderFieldValueCacheControlResponse()
    _header_minimal_bytes = b''
    _header_minimal_markdown = ''

    _header_full = HttpHeaderFieldValueCacheControlResponse(
        max_age=datetime.timedelta(seconds=1),
        s_maxage=datetime.timedelta(seconds=2),
        must_revalidate=True,
        no_cache=True,
        no_store=True,
        public=True,
        private=True,
        no_transform=True,
    )
    _header_full_bytes = b'max-age=1, s-maxage=2, must-revalidate, no-cache, no-store, public, private, no-transform'
    _header_full_upper_case_bytes = b', '.join([
        b'MAX-AGE=1',
        b'S-MAXAGE=2',
        b'MUST-REVALIDATE',
        b'NO-CACHE',
        b'NO-STORE',
        b'PUBLIC',
        b'PRIVATE',
        b'NO-TRANSFORM',
    ])
    _header_full_lower_case_bytes = b', '.join([
        b'max-age=1',
        b's-maxage=2',
        b'must-revalidate',
        b'no-cache',
        b'no-store',
        b'public',
        b'private',
        b'no-transform',
    ])
    _header_minimal_markdown = os.linesep.join([
        '* Max Age: n/a',
        '* S Maxage: n/a',
        '* Must Revalidate: no',
        '* Proxy Revalidate: no',
        '* No Cache: no',
        '* No Store: no',
        '* Public: no',
        '* Private: no',
        '* No Transform: no',
        '',
    ])


class TestContentSecurityPolicySourceHash(unittest.TestCase):
    def test_error_wrong_prefix(self):
        with self.assertRaises(InvalidType):
            ContentSecurityPolicySourceHash.parse_exact_size(b'notavalidhashalgorithm')

    def test_parse(self):
        self.assertEqual(
            ContentSecurityPolicySourceHash.parse_exact_size(b'sha256-bGlnaHQgd29yay4='),
            ContentSecurityPolicySourceHash(Hash.SHA2_256, bytearray(b'light work.'))
        )

    def test_compose(self):
        self.assertEqual(
            ContentSecurityPolicySourceHash(Hash.SHA2_256, bytearray(b'light work.')).compose(),
            b'sha256-bGlnaHQgd29yay4='
        )


class TestContentSecurityPolicySourceHost(unittest.TestCase):
    def test_parse(self):
        self.assertEqual(
            ContentSecurityPolicySourceHost.parse_exact_size(b'http://example.com'),
            ContentSecurityPolicySourceHost('http://example.com')
        )

    def test_compose(self):
        self.assertEqual(
            ContentSecurityPolicySourceHost('http://example.com').compose(),
            b'http://example.com'
        )


class TestContentSecurityPolicySourceNonce(unittest.TestCase):
    def test_error_wrong_prefix(self):
        with self.assertRaises(InvalidType):
            ContentSecurityPolicySourceNonce.parse_exact_size(b'bGlnaHQgd29yay4=')

    def test_parse(self):
        self.assertEqual(
            ContentSecurityPolicySourceNonce.parse_exact_size(b'nonce-bGlnaHQgd29yay4='),
            ContentSecurityPolicySourceNonce(bytearray(b'light work.'))
        )

    def test_compose(self):
        self.assertEqual(
            ContentSecurityPolicySourceNonce(bytearray(b'light work.')).compose(),
            b'nonce-bGlnaHQgd29yay4='
        )


class TestContentSecurityPolicySourceScheme(unittest.TestCase):
    def test_parse(self):
        self.assertEqual(
            ContentSecurityPolicySourceScheme.parse_exact_size(b'http:'),
            ContentSecurityPolicySourceScheme('http')
        )

    def test_compose(self):
        self.assertEqual(
            ContentSecurityPolicySourceScheme('http').compose(),
            b'http:'
        )


class TestContentSecurityPolicyDirectivesFetch(
        TestCasesBasesHttpHeader.MinimalHeader,
        TestCasesBasesHttpHeader.FullHeader):

    _header_minimal = ContentSecurityPolicyDirectiveDefaultSrc([ContentSecurityPolicySourceKeyword.SELF])
    _header_minimal_bytes = ' '.join([
        'default-src',
        '\'self\'',
    ]).encode('ascii')
    _header_minimal_markdown = os.linesep.join([
        '* Type: default-src',
        '* Value:',
        '    1.',
        '        * Type: KEYWORD',
        '        * Value: \'self\'',
        '',
    ])
    _header_full = ContentSecurityPolicyDirectiveDefaultSrc([
        ContentSecurityPolicySourceKeyword.SELF,
        ContentSecurityPolicySourceScheme('http'),
        ContentSecurityPolicySourceHost('https://example.com'),
        ContentSecurityPolicySourceNonce(bytearray(b'light work.')),
        ContentSecurityPolicySourceHash(Hash.SHA2_256, bytearray(b'light work.')),
    ])
    _header_full_bytes = ' '.join([
        'default-src',
        '\'self\'',
        'http:',
        'https://example.com',
        'nonce-bGlnaHQgd29yay4=',
        'sha256-bGlnaHQgd29yay4=',
    ]).encode('ascii')

    def test_min_source_length(self):
        with self.assertRaises(InvalidValue) as context_manager:
            ContentSecurityPolicyDirectiveDefaultSrc.parse_exact_size(b'default-src')
        self.assertEqual(context_manager.exception.value, b'')

    def test_error_invalid_value_type(self):
        with self.assertRaises(InvalidValue):
            ContentSecurityPolicyDirectiveDefaultSrc([None])


class TestContentSecurityPolicyDirectiveFrameAncestors(
        TestCasesBasesHttpHeader.MinimalHeader,
        TestCasesBasesHttpHeader.FullHeader):

    _header_minimal = ContentSecurityPolicyDirectiveFrameAncestors([ContentSecurityPolicySourceKeyword.SELF])
    _header_minimal_bytes = ' '.join([
        'frame-ancestors',
        '\'self\'',
    ]).encode('ascii')
    _header_minimal_markdown = os.linesep.join([
        '* Type: frame-ancestors',
        '* Value:',
        '    1.',
        '        * Type: KEYWORD',
        '        * Value: \'self\'',
        '',
    ])
    _header_full = ContentSecurityPolicyDirectiveFrameAncestors([
        ContentSecurityPolicySourceKeyword.SELF,
        ContentSecurityPolicySourceScheme('http'),
        ContentSecurityPolicySourceHost('https://example.com'),
    ])
    _header_full_bytes = ' '.join([
        'frame-ancestors',
        '\'self\'',
        'http:',
        'https://example.com',
    ]).encode('ascii')

    def test_error_invalid_value_type(self):
        with self.assertRaises(InvalidValue):
            ContentSecurityPolicyDirectiveFrameAncestors([
                ContentSecurityPolicySourceKeyword(ContentSecurityPolicySourceKeyword.REPORT_SAMPLE)
            ])

        with self.assertRaises(InvalidValue):
            ContentSecurityPolicyDirectiveFrameAncestors([
                ContentSecurityPolicySourceNonce(bytearray(b'light work.'))
            ])

        with self.assertRaises(InvalidValue):
            ContentSecurityPolicyDirectiveFrameAncestors([
                ContentSecurityPolicySourceHash(Hash.SHA2_256, bytearray(b'light work.'))
            ])


class TestContentSecurityPolicyDirectiveWebrtc(
        TestCasesBasesHttpHeader.MinimalHeader):

    _header_minimal = ContentSecurityPolicyDirectiveWebrtc(ContentSecurityPolicyWebRtcType.ALLOW)
    _header_minimal_bytes = ' '.join([
        'webrtc',
        '\'allow\'',
    ]).encode('ascii')
    _header_minimal_markdown = '\'allow\''

    def test_error_invalid_value_type(self):
        with self.assertRaises(InvalidValue) as context_manager:
            ContentSecurityPolicyDirectiveWebrtc('not-a-webrtc-type')
        self.assertEqual(context_manager.exception.value, 'not-a-webrtc-type')


class TestContentSecurityPolicyDirectiveRequireTrustedTypesFor(
        TestCasesBasesHttpHeader.MinimalHeader):

    _header_minimal = ContentSecurityPolicyDirectiveRequireTrustedTypesFor([
        ContentSecurityPolicyTrustedTypeSinkGroup.SCRIPT
    ])
    _header_minimal_bytes = ' '.join([
        'require-trusted-types-for',
        '\'script\'',
    ]).encode('ascii')
    _header_minimal_markdown = os.linesep.join([
        '* Type: require-trusted-types-for',
        '* Sink Groups:',
        '    1. \'script\'',
        '',
    ])


class TestContentSecurityPolicyDirectiveReportUri(
        TestCasesBasesHttpHeader.MinimalHeader,
        TestCasesBasesHttpHeader.FullHeader):

    _header_minimal = ContentSecurityPolicyDirectiveReportUri(['http://example.com'])
    _header_minimal_bytes = ' '.join([
        'report-uri',
        'http://example.com',
    ]).encode('ascii')
    _header_minimal_markdown = os.linesep.join([
        '* Type: report-uri',
        '* URI references:',
        '    1. http://example.com',
        '',
    ])
    _header_full = ContentSecurityPolicyDirectiveReportUri(['http://example.com/1', 'http://example.com/2'])
    _header_full_bytes = ' '.join([
        'report-uri',
        'http://example.com/1',
        'http://example.com/2',
    ]).encode('ascii')

    def test_min_reference_length(self):
        with self.assertRaises(InvalidValue) as context_manager:
            ContentSecurityPolicyDirectiveReportUri.parse_exact_size(b'report-uri')
        self.assertEqual(context_manager.exception.value, b'')


class TestContentSecurityPolicyDirectiveReportTo(
        TestCasesBasesHttpHeader.MinimalHeader):

    _header_minimal = ContentSecurityPolicyDirectiveReportTo('token')
    _header_minimal_bytes = ' '.join([
        'report-to',
        'token',
    ]).encode('ascii')
    _header_minimal_markdown = os.linesep.join([
        '* Token: token',
        '',
    ])


class TestContentSecurityPolicyDirectiveSandbox(
        TestCasesBasesHttpHeader.MinimalHeader,
        TestCasesBasesHttpHeader.FullHeader):

    _header_minimal = ContentSecurityPolicyDirectiveSandbox(['token'])
    _header_minimal_bytes = ' '.join([
        'sandbox',
        'token',
    ]).encode('ascii')
    _header_minimal_markdown = os.linesep.join([
        '* Type: sandbox',
        '* Tokens:',
        '    1. token',
        '',
    ])
    _header_full = ContentSecurityPolicyDirectiveSandbox(['token1', 'token2'])
    _header_full_bytes = ' '.join([
        'sandbox',
        'token1',
        'token2',
    ]).encode('ascii')


class TestContentSecurityPolicyDirectivePluginTypes(
        TestCasesBasesHttpHeader.MinimalHeader,
        TestCasesBasesHttpHeader.FullHeader):

    _header_minimal = ContentSecurityPolicyDirectivePluginTypes([
        FieldValueMimeType('html', MimeTypeRegistry.TEXT)
    ])
    _header_minimal_bytes = ' '.join([
        'plugin-types',
        'text/html',
    ]).encode('ascii')
    _header_minimal_markdown = os.linesep.join([
        '* Type: plugin-types',
        '* MIME Types:',
        '    1.',
        '        * Type: html',
        '        * Registry: TEXT',
        '',
    ])
    _header_full = ContentSecurityPolicyDirectivePluginTypes([
        FieldValueMimeType('html', MimeTypeRegistry.TEXT),
        FieldValueMimeType('csv', MimeTypeRegistry.TEXT),
    ])
    _header_full_bytes = ' '.join([
        'plugin-types',
        'text/html',
        'text/csv',
    ]).encode('ascii')


class TestContentSecurityPolicyDirectiveNoValue(
        TestCasesBasesHttpHeader.MinimalHeader):

    _header_minimal = ContentSecurityPolicyDirectiveBlockAllMixedContent()
    _header_minimal_bytes = b'block-all-mixed-content'
    _header_minimal_markdown = os.linesep.join([
        '* Type: block-all-mixed-content',
        '* Value: n/a',
        '',
    ])


class TestHttpHeaderFieldValueContentSecurityPolicy(
        TestCasesBasesHttpHeader.MinimalHeader,
        TestCasesBasesHttpHeader.FullHeader):

    _header_minimal = HttpHeaderFieldValueContentSecurityPolicy([
        ContentSecurityPolicyDirectiveDefaultSrc([
            ContentSecurityPolicySourceKeyword.SELF,
            ContentSecurityPolicySourceHash(Hash.SHA2_256, bytearray(b'light work.')),
            ContentSecurityPolicySourceNonce(bytearray(b'light work.')),
            ContentSecurityPolicySourceScheme('http'),
            ContentSecurityPolicySourceHost('http://example.com'),
        ])
    ])
    _header_minimal_bytes = ' '.join([
        'default-src',
        '\'self\'',
        'sha256-bGlnaHQgd29yay4=',
        'nonce-bGlnaHQgd29yay4=',
        'http:',
        'http://example.com',
    ]).encode('ascii')
    _header_minimal_markdown = os.linesep.join([
        '* Directives:',
        '    1.',
        '        * Type: default-src',
        '        * Value:',
        '            1.',
        '                * Type: KEYWORD',
        '                * Value: \'self\'',
        '            2.',
        '                * Type: HASH',
        '                * Value:',
        '                    * Hash Algorithm: SHA-256',
        '                    * Hash Value: bGlnaHQgd29yay4=',
        '            3.',
        '                * Type: NONCE',
        '                * Value: bGlnaHQgd29yay4=',
        '            4.',
        '                * Type: SCHEME',
        '                * Value: http',
        '            5.',
        '                * Type: HOST',
        '                * Value: http://example.com',
        '',
    ])
    _header_full = HttpHeaderFieldValueContentSecurityPolicy([
        ContentSecurityPolicyDirectiveChildSrc([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveConnectSrc([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveDefaultSrc([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveFontSrc([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveFrameSrc([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveImgSrc([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveManifestSrc([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveMediaSrc([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveObjectSrc([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectivePrefetchSrc([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveScriptSrc([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveScriptSrcElem([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveScriptSrcAttr([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveStyleSrc([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveStyleSrcElem([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveStyleSrcAttr([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveWebrtc(
            ContentSecurityPolicyWebRtcType.ALLOW,
        ),
        ContentSecurityPolicyDirectiveWorkerSrc([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveBaseUri([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveSandbox([
            'token'
        ]),
        ContentSecurityPolicyDirectiveFormAction([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveFrameAncestors([
            ContentSecurityPolicySourceKeyword.SELF,
        ]),
        ContentSecurityPolicyDirectiveReportUri([
            'http://example.com'
        ]),
        ContentSecurityPolicyDirectiveReportTo(
            'token'
        ),
        ContentSecurityPolicyDirectiveBlockAllMixedContent(),
        ContentSecurityPolicyDirectiveUpgradeInsecureRequests(),
        ContentSecurityPolicyDirectiveReferrer(
            ContentSecurityPolicyReferrerPolicy.NO_REFERRER,
        ),
        ContentSecurityPolicyDirectivePluginTypes([
            FieldValueMimeType('html', MimeTypeRegistry.TEXT)
        ]),
    ])
    _header_full_bytes = '; '.join([
        'child-src \'self\'',
        'connect-src \'self\'',
        'default-src \'self\'',
        'font-src \'self\'',
        'frame-src \'self\'',
        'img-src \'self\'',
        'manifest-src \'self\'',
        'media-src \'self\'',
        'object-src \'self\'',
        'prefetch-src \'self\'',
        'script-src \'self\'',
        'script-src-elem \'self\'',
        'script-src-attr \'self\'',
        'style-src \'self\'',
        'style-src-elem \'self\'',
        'style-src-attr \'self\'',
        'webrtc \'allow\'',
        'worker-src \'self\'',
        'base-uri \'self\'',
        'sandbox token',
        'form-action \'self\'',
        'frame-ancestors \'self\'',
        'report-uri http://example.com',
        'report-to token',
        'block-all-mixed-content',
        'upgrade-insecure-requests',
        'referrer "no-referrer"',
        'plugin-types text/html',
    ]).encode('ascii')


class TestHttpHeaderFieldValueContentType(
        TestCasesBasesHttpHeader.MinimalHeader,
        TestCasesBasesHttpHeader.FullHeader):

    _header_minimal = HttpHeaderFieldValueContentType(
        FieldValueMimeType('html', MimeTypeRegistry.TEXT)
     )
    _header_minimal_bytes = b'text/html'
    _header_minimal_markdown = os.linesep.join([
        '* MIME type:',
        '    * Type: html',
        '    * Registry: TEXT',
        '* Charset: n/a',
        '* Boundary: n/a',
        '',
    ])

    _header_full = HttpHeaderFieldValueContentType(
        FieldValueMimeType('bhttp', MimeTypeRegistry.MESSAGE),
        charset='utf-8',
        boundary='boundary_pattern',
    )
    _header_full_bytes = b'message/bhttp; charset=utf-8; boundary=boundary_pattern'

    def test_error_invalid_parameter(self):
        with self.assertRaises(InvalidValue) as context_manager:
            HttpHeaderFieldValueContentType(
                FieldValueMimeType('html', MimeTypeRegistry.TEXT),
                boundary='pattern',
            )
        self.assertEqual(context_manager.exception.value, 'pattern')

        with self.assertRaises(InvalidValue) as context_manager:
            HttpHeaderFieldValueContentType(
                FieldValueMimeType('bhttp', MimeTypeRegistry.MESSAGE),
            )
        self.assertEqual(context_manager.exception.value, None)


class TestHttpHeaderFieldValueNetworkErrorLogging(
        TestCasesBasesHttpHeader.MinimalHeader,
        TestCasesBasesHttpHeader.FullHeader):
    _header_minimal = HttpHeaderFieldValueNetworkErrorLogging(
        report_to="network-errors",
        max_age=1,
    )
    _header_minimal_bytes = b'{"report_to": "network-errors", "max_age": 1}'
    _header_minimal_markdown = os.linesep.join([
        '* Report To: network-errors',
        '* Max Age: 0:00:01',
        '* Include Subdomains: n/a',
        '* Success Fraction: n/a',
        '* Failure Fraction: n/a',
        '',
    ])

    _header_full = HttpHeaderFieldValueNetworkErrorLogging(
        report_to="network-errors",
        max_age=datetime.timedelta(1),
        include_subdomains=True,
        success_fraction=0.1,
        failure_fraction=0.9,
    )
    _header_full_bytes = b''.join([
        b'{',
        b'"report_to": "network-errors", ',
        b'"max_age": 86400, ',
        b'"include_subdomains": true, ',
        b'"success_fraction": 0.1, ',
        b'"failure_fraction": 0.9',
        b'}',
    ])


class TestHttpHeaderFieldValuePragma(
        TestCasesBasesHttpHeader.FullHeader,
        TestCasesBasesHttpHeader.CaseInsensitiveHeader):
    _header_full = HttpHeaderFieldValuePragma(HttpHeaderPragma.NO_CACHE)
    _header_full_bytes = b'no-cache'
    _header_full_upper_case_bytes = b'NO-CACHE'
    _header_full_lower_case_bytes = b'no-cache'


class TestHttpHeaderFieldValueSTS(
        TestCasesBasesHttpHeader.MinimalHeader,
        TestCasesBasesHttpHeader.FullHeader,
        TestCasesBasesHttpHeader.CaseInsensitiveHeader):
    _header_minimal = HttpHeaderFieldValueSTS(max_age=datetime.timedelta(seconds=1))
    _header_minimal_bytes = b'max-age=1'
    _header_minimal_markdown = os.linesep.join([
        '* Max Age: 0:00:01',
        '* Include Subdomains: no',
        '* Preload: no',
        '',
    ])

    _header_full = HttpHeaderFieldValueSTS(max_age=datetime.timedelta(seconds=1), include_subdomains=True, preload=True)
    _header_full_bytes = b'max-age=1; includeSubDomains; preload'
    _header_full_upper_case_bytes = b'MAX-AGE=1; INCLUDESUBDOMAINS; PRELOAD'
    _header_full_lower_case_bytes = b'max-age=1; includesubdomains; preload'


class TestHttpHeaderFieldValueExpectStaple(
        TestCasesBasesHttpHeader.MinimalHeader,
        TestCasesBasesHttpHeader.FullHeader,
        TestCasesBasesHttpHeader.CaseInsensitiveHeader):
    _header_minimal = HttpHeaderFieldValueExpectStaple(max_age=datetime.timedelta(seconds=1))
    _header_minimal_bytes = b'max-age=1'
    _header_minimal_markdown = os.linesep.join([
        '* Max Age: 0:00:01',
        '* Include Subdomains: no',
        '* Preload: no',
        '* Report Uri: n/a',
        '',
    ])

    _header_full = HttpHeaderFieldValueExpectStaple(
        max_age=datetime.timedelta(seconds=1),
        include_subdomains=True,
        preload=True,
        report_uri="http://example.com"
    )
    _header_full_bytes = b'max-age=1; includeSubDomains; preload; report-uri="http://example.com"'
    _header_full_upper_case_bytes = b'MAX-AGE=1; INCLUDESUBDOMAINS; PRELOAD; REPORT-URI="http://example.com"'
    _header_full_lower_case_bytes = b'max-age=1; includesubdomains; preload; report-uri="http://example.com"'


class TestHttpHeaderFieldValueExpectCT(
        TestCasesBasesHttpHeader.MinimalHeader,
        TestCasesBasesHttpHeader.FullHeader,
        TestCasesBasesHttpHeader.CaseInsensitiveHeader):
    _header_minimal = HttpHeaderFieldValueExpectCT(max_age=datetime.timedelta(seconds=1))
    _header_minimal_bytes = b'max-age=1'
    _header_minimal_markdown = os.linesep.join([
        '* Max Age: 0:00:01',
        '* Enforce: no',
        '* Report Uri: n/a',
        '',
    ])

    _header_full = HttpHeaderFieldValueExpectCT(
        max_age=datetime.timedelta(seconds=1),
        enforce=True,
        report_uri='http://example.com'
    )
    _header_full_bytes = b'max-age=1, enforce, report-uri="http://example.com"'
    _header_full_upper_case_bytes = b'MAX-AGE=1, ENFORCE, REPORT-URI="http://example.com"'
    _header_full_lower_case_bytes = b'max-age=1, enforce, report-uri="http://example.com"'


class TestHttpHeaderFieldValuePublicKeyPinning(
        TestCasesBasesHttpHeader.MinimalHeader,
        TestCasesBasesHttpHeader.FullHeader,
        TestCasesBasesHttpHeader.CaseInsensitiveHeader):
    _header_minimal = HttpHeaderFieldValuePublicKeyPinning(
        pin_sha256='cGluLXNoYTI1Ng==',
        max_age=datetime.timedelta(seconds=1),
    )
    _header_minimal_bytes = b'pin-sha256="cGluLXNoYTI1Ng=="; max-age=1'
    _header_minimal_markdown = os.linesep.join([
        '* Pin (SHA-256): cGluLXNoYTI1Ng==',
        '* Max Age: 0:00:01',
        '* Include Subdomains: no',
        '* Report Uri: n/a',
        '',
    ])

    _header_full = HttpHeaderFieldValuePublicKeyPinning(
        pin_sha256='cGluLXNoYTI1Ng==',
        max_age=datetime.timedelta(seconds=1),
        include_subdomains=True,
        report_uri='http://example.com'
    )
    _header_full_bytes = b'; '.join([
        b'pin-sha256="cGluLXNoYTI1Ng=="',
        b'max-age=1',
        b'includeSubDomains',
        b'report-uri="http://example.com"',
    ])
    _header_full_upper_case_bytes = b'; '.join([
        b'PIN-SHA256="cGluLXNoYTI1Ng=="',
        b'MAX-AGE=1',
        b'INCLUDESUBDOMAINS',
        b'REPORT-URI="http://example.com"',
    ])
    _header_full_lower_case_bytes = b'; '.join([
        b'pin-sha256="cGluLXNoYTI1Ng=="',
        b'max-age=1',
        b'includesubdomains',
        b'report-uri="http://example.com"',
    ])


class TestHttpHeaderFieldValueSetCookie(TestCasesBasesHttpHeader.MinimalHeader, TestCasesBasesHttpHeader.FullHeader):
    _header_minimal = HttpHeaderFieldValueSetCookie('name', 'value')
    _header_minimal_bytes = b'name=value'
    _header_minimal_markdown = os.linesep.join([
        '* Name: name',
        '* Value: value',
        '* Expires: n/a',
        '* Max Age: n/a',
        '* Domain: n/a',
        '* Path: n/a',
        '* Secure: no',
        '* Http Only: no',
        '* Same Site: n/a',
        ''
    ])

    _header_full = HttpHeaderFieldValueSetCookie(
        name='name', value='value',
        expires=datetime.datetime.fromtimestamp(0, datetime.timezone.utc),
        max_age=datetime.timedelta(seconds=1),
        domain='example.com',
        path='/',
        secure=True,
        http_only=True,
        same_site=HttpHeaderSetCookieComponentSameSite.LAX,
    )
    _header_full_bytes = b'; '.join([
        b'name=value',
        b'expires=Thu, 01 Jan 1970 00:00:00 GMT',
        b'max-age=1',
        b'Domain=example.com',
        b'Path=/',
        b'Secure',
        b'HttpOnly',
        b'SameSite=Lax',
    ])


class TestHttpHeaderFieldValueXContentTypeOptions(TestCasesBasesHttpHeader.FullHeader):
    _header_full = HttpHeaderFieldValueXContentTypeOptions(HttpHeaderXContentTypeOptions.NOSNIFF)
    _header_full_bytes = b'nosniff'


class TestHttpHeaderFieldValueXFrameOptions(TestCasesBasesHttpHeader.FullHeader):
    _header_full = HttpHeaderFieldValueXFrameOptions(HttpHeaderXFrameOptions.SAMEORIGIN)
    _header_full_bytes = b'SAMEORIGIN'


class TestHttpHeaderFieldValueXXSSProtection(TestCasesBasesHttpHeader.FullHeader):
    _header_full = HttpHeaderFieldValueXXSSProtection(
        HttpHeaderXXSSProtectionState.ENABLED,
        HttpHeaderXXSSProtectionMode.BLOCK,
        'http://example.com'
    )
    _header_full_bytes = b'1; mode=block; report=http://example.com'


class TestHttpHeaderFieldValueReferrerPolicy(TestCasesBasesHttpHeader.FullHeader):
    _header_full = HttpHeaderFieldValueReferrerPolicy(HttpHeaderReferrerPolicy.SAME_ORIGIN)
    _header_full_bytes = b'same-origin'


class TestHttpHeaderFieldName(unittest.TestCase):
    def test_markdown(self):
        self.assertEqual(
            HttpHeaderFieldName.STRICT_TRANSPORT_SECURITY.value.as_markdown(),
            'Strict-Transport-Security',
        )

    def test_from_name(self):
        with self.assertRaises(InvalidValue) as context_manager:
            HttpHeaderFieldName.from_name('non-existing-name')
        self.assertEqual(context_manager.exception.value, 'non-existing-name')

        self.assertEqual(
            HttpHeaderFieldName.from_name('strict-transport-security'),
            HttpHeaderFieldName.STRICT_TRANSPORT_SECURITY
        )
        self.assertEqual(
            HttpHeaderFieldName.from_name('STRICT-TRANSPORT-SECURITY'),
            HttpHeaderFieldName.STRICT_TRANSPORT_SECURITY
        )
        self.assertEqual(
            HttpHeaderFieldName.from_name('Strict-Transport-Security'),
            HttpHeaderFieldName.STRICT_TRANSPORT_SECURITY
        )


class TestHttpHeaderFields(unittest.TestCase):
    def setUp(self):
        self.headers_bytes = b'\r\n'.join([
            b'Age: 1',
            b'Cache-Control: no-cache',
            b'Content-Type: text/html',
            b'Date: Thu, 01 Jan 1970 00:00:00 GMT',
            b'ETag: 12345678',
            b'Expect-CT: max-age=1',
            b'Expect-Staple: max-age=1',
            b'Expires: Thu, 01 Jan 1970 00:00:00 GMT',
            b'Last-Modified: Thu, 01 Jan 1970 00:00:00 GMT',
            b'NEL: {"report_to": "network-errors", "max_age": 1}',
            b'Pragma: no-cache',
            b'Public-Key-Pinning: pin-sha256="cGluLXNoYTI1Ng=="; max-age=1',
            b'Referrer-Policy: origin',
            b'Server: server',
            b'Set-Cookie: name=value',
            b'Strict-Transport-Security: max-age=1',
            b'X-Unparsed: Value',
            b'X-Content-Type-Options: nosniff',
            b'X-Frame-Options: SAMEORIGIN',
            b'X-XSS-Protection: 1',
            b'Content-Security-Policy: default-src \'self\'',
            b'Content-Security-Policy-Report-Only: default-src \'self\'',
            b'X-Content-Security-Policy: default-src \'self\'',
            b'',
            b'',
        ])
        self.headers = HttpHeaderFields([
            HttpHeaderFieldAge(datetime.timedelta(seconds=1)),
            HttpHeaderFieldCacheControlResponse(HttpHeaderFieldValueCacheControlResponse(no_cache=True)),
            HttpHeaderFieldContentType(FieldValueMimeType('html', MimeTypeRegistry.TEXT)),
            HttpHeaderFieldDate(datetime.datetime.fromtimestamp(0, datetime.timezone.utc)),
            HttpHeaderFieldETag('12345678'),
            HttpHeaderFieldExpectCT(HttpHeaderFieldValueExpectCT(datetime.timedelta(seconds=1))),
            HttpHeaderFieldExpectStaple(HttpHeaderFieldValueExpectStaple(datetime.timedelta(seconds=1))),
            HttpHeaderFieldExpires(datetime.datetime.fromtimestamp(0, datetime.timezone.utc)),
            HttpHeaderFieldLastModified(datetime.datetime.fromtimestamp(0, datetime.timezone.utc)),
            HttpHeaderFieldNetworkErrorLogging(HttpHeaderFieldValueNetworkErrorLogging(
                report_to="network-errors", max_age=1
            )),
            HttpHeaderFieldPragma(HttpHeaderPragma.NO_CACHE),
            HttpHeaderFieldPublicKeyPinning(HttpHeaderFieldValuePublicKeyPinning(
                pin_sha256='cGluLXNoYTI1Ng==',
                max_age=datetime.timedelta(seconds=1),
            )),
            HttpHeaderFieldReferrerPolicy(HttpHeaderReferrerPolicy.ORIGIN),
            HttpHeaderFieldServer('server'),
            HttpHeaderFieldSetCookie(HttpHeaderFieldValueSetCookie('name', 'value')),
            HttpHeaderFieldSTS(HttpHeaderFieldValueSTS(datetime.timedelta(seconds=1))),
            HttpHeaderFieldUnparsed('X-Unparsed', 'Value'),
            HttpHeaderFieldXContentTypeOptions(HttpHeaderXContentTypeOptions.NOSNIFF),
            HttpHeaderFieldXFrameOptions(HttpHeaderXFrameOptions.SAMEORIGIN),
            HttpHeaderFieldXXSSProtection(HttpHeaderXXSSProtectionState.ENABLED),
            HttpHeaderFieldContentSecurityPolicy([
                ContentSecurityPolicyDirectiveDefaultSrc([ContentSecurityPolicySourceKeyword.SELF])
            ]),
            HttpHeaderFieldContentSecurityPolicyReportOnly([
                ContentSecurityPolicyDirectiveDefaultSrc([ContentSecurityPolicySourceKeyword.SELF])
            ]),
            HttpHeaderFieldXContentSecurityPolicy([
                ContentSecurityPolicyDirectiveDefaultSrc([ContentSecurityPolicySourceKeyword.SELF])
            ]),
        ])

    def test_parse(self):
        self.assertEqual(
            HttpHeaderFields.parse_exact_size(self.headers_bytes),
            self.headers
        )

    def test_compose(self):
        self.assertEqual(self.headers.compose(), self.headers_bytes)

    def test_markdown(self):
        self.assertEqual(self.headers.as_markdown(), os.linesep.join([
            '1.',
            '    * Name: Age',
            '    * Value: 1',
            '2.',
            '    * Name: Cache-Control',
            '    * Value:',
            '        * Max Age: n/a',
            '        * S Maxage: n/a',
            '        * Must Revalidate: no',
            '        * Proxy Revalidate: no',
            '        * No Cache: yes',
            '        * No Store: no',
            '        * Public: no',
            '        * Private: no',
            '        * No Transform: no',
            '3.',
            '    * Name: Content-Type',
            '    * Value:',
            '        * MIME type:',
            '            * Type: html',
            '            * Registry: TEXT',
            '        * Charset: n/a',
            '        * Boundary: n/a',
            '4.',
            '    * Name: Date',
            '    * Value: 1970-01-01 00:00:00+00:00',
            '5.',
            '    * Name: ETag',
            '    * Value: 12345678',
            '6.',
            '    * Name: Expect-CT',
            '    * Value:',
            '        * Max Age: 0:00:01',
            '        * Enforce: no',
            '        * Report Uri: n/a',
            '7.',
            '    * Name: Expect-Staple',
            '    * Value:',
            '        * Max Age: 0:00:01',
            '        * Include Subdomains: no',
            '        * Preload: no',
            '        * Report Uri: n/a',
            '8.',
            '    * Name: Expires',
            '    * Value: 1970-01-01 00:00:00+00:00',
            '9.',
            '    * Name: Last-Modified',
            '    * Value: 1970-01-01 00:00:00+00:00',
            '10.',
            '    * Name: NEL',
            '    * Value:',
            '        * Report To: network-errors',
            '        * Max Age: 0:00:01',
            '        * Include Subdomains: n/a',
            '        * Success Fraction: n/a',
            '        * Failure Fraction: n/a',
            '11.',
            '    * Name: Pragma',
            '    * Value: no-cache',
            '12.',
            '    * Name: Public-Key-Pinning',
            '    * Value:',
            '        * Pin (SHA-256): cGluLXNoYTI1Ng==',
            '        * Max Age: 0:00:01',
            '        * Include Subdomains: no',
            '        * Report Uri: n/a',
            '13.',
            '    * Name: Referrer-Policy',
            '    * Value: origin',
            '14.',
            '    * Name: Server',
            '    * Value: server',
            '15.',
            '    * Name: Set-Cookie',
            '    * Value:',
            '        * Name: name',
            '        * Value: value',
            '        * Expires: n/a',
            '        * Max Age: n/a',
            '        * Domain: n/a',
            '        * Path: n/a',
            '        * Secure: no',
            '        * Http Only: no',
            '        * Same Site: n/a',
            '16.',
            '    * Name: Strict-Transport-Security',
            '    * Value:',
            '        * Max Age: 0:00:01',
            '        * Include Subdomains: no',
            '        * Preload: no',
            '17.',
            '    * Name: X-Unparsed',
            '    * Value: Value',
            '18.',
            '    * Name: X-Content-Type-Options',
            '    * Value: nosniff',
            '19.',
            '    * Name: X-Frame-Options',
            '    * Value: SAMEORIGIN',
            '20.',
            '    * Name: X-XSS-Protection',
            '    * Value:',
            '        * State: enabled',
            '        * Mode: n/a',
            '        * Report: n/a',
            '21.',
            '    * Name: Content-Security-Policy',
            '    * Value:',
            '        * Directives:',
            '            1.',
            '                * Type: default-src',
            '                * Value:',
            '                    1.',
            '                        * Type: KEYWORD',
            '                        * Value: \'self\'',
            '22.',
            '    * Name: Content-Security-Policy-Report-Only',
            '    * Value:',
            '        * Directives:',
            '            1.',
            '                * Type: default-src',
            '                * Value:',
            '                    1.',
            '                        * Type: KEYWORD',
            '                        * Value: \'self\'',
            '23.',
            '    * Name: X-Content-Security-Policy',
            '    * Value:',
            '        * Directives:',
            '            1.',
            '                * Type: default-src',
            '                * Value:',
            '                    1.',
            '                        * Type: KEYWORD',
            '                        * Value: \'self\'',
            '',
        ]))


class TestHttpHeaderFieldUnparsed(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            HttpHeaderFieldUnparsed.parse_immutable(b'name: value')
        self.assertEqual(context_manager.exception.value, b'value')

        with self.assertRaises(InvalidValue) as context_manager:
            HttpHeaderFieldUnparsed.parse_immutable(b'name value')
        self.assertEqual(context_manager.exception.value, b'')

    def test_parse(self):
        parsable = b'name: value\r\n'
        header, parsed_length = HttpHeaderFieldUnparsed.parse_immutable(parsable)
        self.assertEqual(header.name, 'name')
        self.assertEqual(header.value, 'value')
        self.assertEqual(parsable[parsed_length:], b'\r\n')

        parsable = b'name:  value\r\n'
        header, parsed_length = HttpHeaderFieldUnparsed.parse_immutable(parsable)
        self.assertEqual(header.name, 'name')
        self.assertEqual(header.value, 'value')
        self.assertEqual(parsable[parsed_length:], b'\r\n')

    def test_compose(self):
        header = HttpHeaderFieldUnparsed('name', 'value')
        self.assertEqual(header.compose(), b'name: value')

    def test_markdown(self):
        header = HttpHeaderFieldUnparsed('name', 'value')
        self.assertEqual(header.as_markdown(), os.linesep.join([
            '* Name: name',
            '* Value: value',
            ''
        ]))



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines

import abc
import calendar
import collections
import datetime
import enum
import random
import attr


from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.tls.algorithm import SslCipherKind, TlsCipherSuite, TlsCipherSuiteExtension, TlsCompressionMethod

from cryptoparser.common.base import (
    OneByteEnumParsable,
    Opaque,
    OpaqueParam,
    VariantParsable,
    Vector,
    VectorEnumCodeNumeric,
    VectorParamEnumCodeNumeric,
    VectorParamNumeric,
    VectorParamParsable,
    VectorParsable,
)
from cryptoparser.common.exception import NotEnoughData, InvalidType
from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary

from cryptoparser.tls.extension import (
    TlsCertificateStatusType,
    TlsExtensionType,
    TlsExtensionsClient,
    TlsExtensionsServer,
    TlsSignatureAndHashAlgorithmVector,
)
from cryptoparser.tls.grease import TlsInvalidType, TlsInvalidTypeOneByte, TlsInvalidTypeTwoByte
from cryptoparser.tls.version import TlsProtocolVersion, TlsVersion
from cryptoparser.tls.ciphersuite import SslCipherKindFactory, TlsCipherSuiteFactory


class TlsContentType(enum.IntEnum):
    CHANGE_CIPHER_SPEC = 0x14
    ALERT = 0x15
    HANDSHAKE = 0x16
    APPLICATION_DATA = 0x17
    HEARTBEAT = 0x18


@attr.s
class SubprotocolParser():
    _subprotocol_type = attr.ib(validator=attr.validators.instance_of(enum.IntEnum))

    @classmethod
    @abc.abstractmethod
    def _get_subprotocol_parsers(cls):
        raise NotImplementedError()

    @classmethod
    def register_subprotocol_parser(cls, subprotocol_type, parsable_class):
        subprotocol_parsers = cls._get_subprotocol_parsers()
        subprotocol_parsers[subprotocol_type] = parsable_class

    def parse(self, parsable):
        subprotocol_parsers = self._get_subprotocol_parsers()

        if self._subprotocol_type in subprotocol_parsers:
            parsed_object, parsed_length = subprotocol_parsers[self._subprotocol_type].parse_immutable(parsable)
            return parsed_object, parsed_length

        raise InvalidValue(self._subprotocol_type, TlsSubprotocolMessageBase)


class TlsSubprotocolMessageBase(ParsableBase):
    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class TlsAlertLevel(enum.IntEnum):
    WARNING = 0x01
    FATAL = 0x02


class TlsAlertDescription(enum.IntEnum):
    CLOSE_NOTIFY = 0x00
    UNEXPECTED_MESSAGE = 0x0a
    BAD_RECORD_MAC = 0x14
    RECORD_OVERFLOW = 0x16
    HANDSHAKE_FAILURE = 0x28
    BAD_CERTIFICATE = 0x2a
    UNSUPPORTED_CERTIFICATE = 0x2b
    CERTIFICATE_REVOKED = 0x2c
    CERTIFICATE_EXPIRED = 0x2d
    CERTIFICATE_UNKNOWN = 0x2e
    ILLEGAL_PARAMETER = 0x2f
    UNKNOWN_CA = 0x30
    ACCESS_DENIED = 0x30
    DECODE_ERROR = 0x32
    DECRYPT_ERROR = 0x33
    PROTOCOL_VERSION = 0x46
    INSUFFICIENT_SECURITY = 0x47
    INTERNAL_ERROR = 0x50
    INAPPROPRIATE_FALLBACK = 0x56
    USER_CANCELED = 0x5a
    MISSING_EXTENSION = 0x6d
    UNSUPPORTED_EXTENSION = 0x6e
    CERTIFICATE_UNOBTAINABLE = 0x6f
    UNRECOGNIZED_NAME = 0x70
    BAD_CERTIFICATE_STATUS_RESPONSE = 0x71
    BAD_CERTIFICATE_HASH_VALUE = 0x72
    UNKNOWN_PSK_IDENTITY = 0x73
    CERTIFICATE_REQUIRED = 0x74
    NO_APPLICATION_PROTOCOL = 0x78


@attr.s
class TlsAlertMessage(TlsSubprotocolMessageBase):
    _SIZE = 2

    level = attr.ib()
    description = attr.ib()

    @classmethod
    def _parse(cls, parsable):
        if len(parsable) < cls._SIZE:
            raise NotEnoughData(cls._SIZE - len(parsable))

        parser = ParserBinary(parsable)

        parser.parse_numeric('level', 1)
        parser.parse_numeric('description', 1)

        return TlsAlertMessage(parser['level'], parser['description']), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.level, 1)
        composer.compose_numeric(self.description, 1)

        return composer.composed_bytes

    @level.validator
    def _validator_level(self, attribute, value):  # pylint: disable=unused-argument
        try:
            self.level = TlsAlertLevel(value)
        except ValueError as e:
            raise InvalidValue(value, TlsAlertLevel, 'level') from e

    @description.validator
    def _validator_description(self, attribute, value):  # pylint: disable=unused-argument
        try:
            self.description = TlsAlertDescription(value)
        except ValueError as e:
            raise InvalidValue(value, TlsAlertDescription) from e


class TlsChangeCipherSpecType(enum.IntEnum):
    CHANGE_CIPHER_SPEC = 0x01


@attr.s
class TlsChangeCipherSpecMessage(TlsSubprotocolMessageBase):
    _change_cipher_spec_type = attr.ib(
        default=TlsChangeCipherSpecType.CHANGE_CIPHER_SPEC,
        validator=attr.validators.in_(TlsChangeCipherSpecType)
    )

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_numeric('change_cipher_spec_type', 1, TlsChangeCipherSpecType)

        return TlsChangeCipherSpecMessage(parser['change_cipher_spec_type']), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self._change_cipher_spec_type, 1)

        return composer.composed_bytes


@attr.s
class TlsApplicationDataMessage(TlsSubprotocolMessageBase):
    data = attr.ib(attr.validators.instance_of(bytearray))

    @classmethod
    def _parse(cls, parsable):
        return TlsApplicationDataMessage(parsable), len(parsable)

    def compose(self):
        return self.data


class TlsHandshakeType(enum.IntEnum):
    HELLO_REQUEST = 0x00
    CLIENT_HELLO = 0x01
    SERVER_HELLO = 0x02
    HELLO_VERIFY_REQUEST = 0x03
    NEW_SESSION_TICKET = 0x04
    HELLO_RETRY_REQUEST = 0x06
    CERTIFICATE = 0x0b
    SERVER_KEY_EXCHANGE = 0x0c
    CERTIFICATE_REQUEST = 0x0d
    SERVER_HELLO_DONE = 0x0e
    CERTIFICATE_VERIFY = 0x0f
    CLIENT_KEY_EXCHANGE = 0x10
    CLIENT_CERTIFICATE_REQUEST = 0x11
    FINISHED = 0x14
    CLIENT_CERTIFICATE_URL = 0x15
    CERTIFICATE_STATUS = 0x16
    SUPPLEMENTAL_DATA = 0x17
    KEY_UPDATE = 0x18
    COMPRESSED_CERTIFICATE = 0x19
    EKT_KEY = 0x15
    MESSAGE_HASH = 254


@attr.s
class TlsHandshakeMessage(TlsSubprotocolMessageBase):
    """The payload of a handshake record.
    """
    _HEADER_SIZE = 4

    @classmethod
    @abc.abstractmethod
    def get_handshake_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_handshake_header(cls, parsable):
        if len(parsable) < cls._HEADER_SIZE:
            raise NotEnoughData(cls._HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)

        try:
            parser.parse_numeric('handshake_type', 1, TlsHandshakeType)
        except InvalidValue as e:
            raise e

        if parser['handshake_type'] != cls.get_handshake_type():
            raise InvalidType()

        try:
            parser.parse_bytes('payload', 3)
        except NotEnoughData as e:
            raise NotEnoughData(e.bytes_needed) from e

        return parser

    def _compose_header(self, payload_length):
        composer = ComposerBinary()

        composer.compose_numeric(self.get_handshake_type(), 1)
        composer.compose_numeric(payload_length, 3)

        return composer.composed_bytes


class TlsHandshakeHelloRandomBytes(Vector):
    @classmethod
    def _parse(cls, parsable):
        composer = ComposerBinary()
        vector_param = cls.get_param()
        composer.compose_numeric(vector_param.min_byte_num, vector_param.item_num_size)

        vector, parsed_length = super(TlsHandshakeHelloRandomBytes, cls)._parse(composer.composed_bytes + parsable)

        return cls(vector), parsed_length - vector_param.item_num_size

    def compose(self):
        return super().compose()[self.get_param().item_num_size:]

    @classmethod
    def get_param(cls):
        return OpaqueParam(min_byte_num=28, max_byte_num=28)


@attr.s
class TlsHandshakeHelloRandom(ParsableBase):
    time = attr.ib(validator=attr.validators.instance_of(datetime.datetime))
    random = attr.ib(validator=attr.validators.instance_of(TlsHandshakeHelloRandomBytes))

    @time.default
    def _default_time(self):  # pylint: disable=no-self-use
        return datetime.datetime.now(datetime.timezone.utc)

    @random.default
    def _default_random(self):  # pylint: disable=no-self-use
        return TlsHandshakeHelloRandomBytes(
            bytearray.fromhex(f'{random.getrandbits(224):28x}'.zfill(56))
        )

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_timestamp('time', item_size=4)
        parser.parse_parsable('random', TlsHandshakeHelloRandomBytes)

        return TlsHandshakeHelloRandom(parser['time'], parser['random']), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(int(calendar.timegm(self.time.utctimetuple())), 4)
        composer.compose_parsable(self.random)

        return composer.composed_bytes


@attr.s
class TlsHandshakeHello(TlsHandshakeMessage):
    @classmethod
    def _parse_hello_header(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_parsable('protocol_version', TlsProtocolVersion)
        parser.parse_parsable('random', TlsHandshakeHelloRandom)
        parser.parse_parsable('session_id', TlsSessionIdVector)

        return parser

    def _compose_header(self, payload_length):
        composer = ComposerBinary()

        handshake_header_bytes = super()._compose_header(
            payload_length + composer.composed_length
        )

        return handshake_header_bytes + composer.composed_bytes

    @classmethod
    def _parse_extensions(cls, handshake_header_parser, parser, extensions_class):
        if parser.parsed_length >= len(handshake_header_parser['payload']):
            return None

        parser.parse_parsable('extensions', extensions_class)

        return parser

    @staticmethod
    def _compose_extensions(extensions):
        extension_bytes = bytearray()

        for extension in extensions:
            extension_bytes += extension.compose()

        payload_composer = ComposerBinary()
        if extensions:
            payload_composer.compose_numeric(len(extension_bytes), 2)

        return payload_composer.composed_bytes + extension_bytes

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class TlsCipherSuiteVector(VectorParsable):
    @classmethod
    def get_param(cls):
        return VectorParamEnumCodeNumeric(
            item_class=TlsCipherSuiteFactory,
            fallback_class=TlsInvalidTypeTwoByte,
            min_byte_num=2, max_byte_num=2 ** 16 - 2
        )


class TlsCompressionMethodFactory(OneByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return TlsCompressionMethod

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class TlsCompressionMethodVector(VectorEnumCodeNumeric):
    @classmethod
    def get_param(cls):
        return VectorParamEnumCodeNumeric(
            item_class=TlsCompressionMethodFactory,
            fallback_class=TlsInvalidTypeOneByte,
            min_byte_num=1,
            max_byte_num=2 ** 8 - 1,
        )


class TlsSessionIdVector(Vector):
    @classmethod
    def get_param(cls):
        return VectorParamNumeric(item_size=1, min_byte_num=0, max_byte_num=32)


@attr.s
class TlsHandshakeClientHello(TlsHandshakeHello):  # pylint: disable=too-many-instance-attributes
    cipher_suites = attr.ib(
        converter=TlsCipherSuiteVector,
        validator=attr.validators.instance_of(TlsCipherSuiteVector)
    )
    protocol_version = attr.ib(
        default=TlsProtocolVersion(TlsVersion.TLS1_2),
        validator=attr.validators.instance_of(TlsProtocolVersion),
    )
    random = attr.ib(
        default=TlsHandshakeHelloRandom(),
        validator=attr.validators.instance_of(TlsHandshakeHelloRandom),
    )
    session_id = attr.ib(
        default=TlsSessionIdVector(()),
        converter=TlsSessionIdVector,
        validator=attr.validators.instance_of(TlsSessionIdVector),
    )
    compression_methods = attr.ib(
        default=TlsCompressionMethodVector([TlsCompressionMethod.NULL, ]),
        converter=TlsCompressionMethodVector,
        validator=attr.validators.instance_of(TlsCompressionMethodVector),
    )
    extensions = attr.ib(
        default=TlsExtensionsClient(()),
        converter=TlsExtensionsClient,
        validator=attr.validators.instance_of(TlsExtensionsClient)
    )
    fallback_scsv = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    empty_renegotiation_info_scsv = attr.ib(default=True, validator=attr.validators.instance_of(bool))

    @classmethod
    def get_handshake_type(cls):
        return TlsHandshakeType.CLIENT_HELLO

    @classmethod
    def _parse(cls, parsable):

        handshake_header_parser = cls._parse_handshake_header(parsable)

        parser = cls._parse_hello_header(handshake_header_parser['payload'])
        parser.parse_parsable('cipher_suites', TlsCipherSuiteVector)
        parser.parse_parsable('compression_methods', TlsCompressionMethodVector)

        extension_parser = cls._parse_extensions(handshake_header_parser, parser, TlsExtensionsClient)

        cipher_suites = []
        fallback_scsv = False
        empty_renegotiation_info_scsv = False
        for cipher_suite in parser['cipher_suites']:
            if cipher_suite.value.code == TlsCipherSuiteExtension.FALLBACK_SCSV.value.code:
                fallback_scsv = True
            elif cipher_suite.value.code == TlsCipherSuiteExtension.EMPTY_RENEGOTIATION_INFO_SCSV.value.code:
                empty_renegotiation_info_scsv = True
            else:
                cipher_suites.append(cipher_suite)

        return TlsHandshakeClientHello(
            cipher_suites,
            parser['protocol_version'],
            parser['random'],
            parser['session_id'],
            parser['compression_methods'],
            extensions=parser['extensions'] if extension_parser else TlsExtensionsClient([]),
            fallback_scsv=fallback_scsv,
            empty_renegotiation_info_scsv=empty_renegotiation_info_scsv,
        ), handshake_header_parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.protocol_version)
        payload_composer.compose_parsable(self.random)
        payload_composer.compose_parsable(self.session_id)
        if self.fallback_scsv:
            self.cipher_suites.append(TlsCipherSuiteExtension.FALLBACK_SCSV)
        if self.empty_renegotiation_info_scsv:
            self.cipher_suites.append(TlsCipherSuiteExtension.EMPTY_RENEGOTIATION_INFO_SCSV)
        payload_composer.compose_numeric(len(self.cipher_suites) * self.cipher_suites.get_param().item_num_size, 2)
        payload_composer.compose_numeric_array_enum_coded(self.cipher_suites)
        if self.fallback_scsv:
            del self.cipher_suites[-1]
        if self.empty_renegotiation_info_scsv:
            del self.cipher_suites[-1]
        payload_composer.compose_parsable(self.compression_methods)

        extension_bytes = self._compose_extensions(self.extensions)

        header_bytes = self._compose_header(payload_composer.composed_length + len(extension_bytes))

        return header_bytes + payload_composer.composed_bytes + extension_bytes

    def ja3(self):
        parser = ParserBinary(self.protocol_version.compose())
        parser.parse_numeric('tls_protocol_version', 2)

        cipher_suites = [str(cipher_suite.value.code) for cipher_suite in self.cipher_suites]

        extension_types = []
        named_curves = []
        ec_point_formats = []
        for extension in self.extensions:
            if (not isinstance(extension.extension_type, TlsInvalidTypeTwoByte) or
                    extension.extension_type.value.value_type != TlsInvalidType.GREASE):
                extension_types.append(str(extension.extension_type.value.code))

            if extension.extension_type == TlsExtensionType.SUPPORTED_GROUPS:
                named_curves = [
                    str(named_curve.value.code)
                    for named_curve in extension.elliptic_curves
                    if (not isinstance(named_curve, TlsInvalidTypeTwoByte) or
                        named_curve.value.value_type != TlsInvalidType.GREASE)
                ]
            elif extension.extension_type == TlsExtensionType.EC_POINT_FORMATS:
                ec_point_formats = [
                    str(point_format.value.code)
                    for point_format in extension.point_formats
                    if (not isinstance(point_format, TlsInvalidTypeOneByte) or
                        point_format.value.value_type != TlsInvalidType.GREASE)
                ]

        return ','.join([
            str(parser['tls_protocol_version']),
            '-'.join(cipher_suites),
            '-'.join(extension_types),
            '-'.join(named_curves),
            '-'.join(ec_point_formats),
        ])


@attr.s
class TlsHandshakeServerHello(TlsHandshakeHello):
    protocol_version = attr.ib(
        default=TlsProtocolVersion(TlsVersion.TLS1_2),
        validator=attr.validators.instance_of(TlsProtocolVersion),
    )
    random = attr.ib(
        default=TlsHandshakeHelloRandom(),
        validator=attr.validators.instance_of(TlsHandshakeHelloRandom),
    )
    session_id = attr.ib(
        default=TlsSessionIdVector((random.randint(0, 255) for i in range(32))),
        converter=TlsSessionIdVector,
        validator=attr.validators.instance_of(TlsSessionIdVector),
    )
    compression_method = attr.ib(
        default=TlsCompressionMethod.NULL,
        converter=TlsCompressionMethod,
        validator=attr.validators.in_(TlsCompressionMethod),
    )
    cipher_suite = attr.ib(default=None, validator=attr.validators.in_(TlsCipherSuite))
    extensions = attr.ib(
        default=TlsExtensionsServer([]),
        converter=TlsExtensionsServer,
        validator=attr.validators.instance_of(TlsExtensionsServer)
    )

    @classmethod
    def get_handshake_type(cls):
        return TlsHandshakeType.SERVER_HELLO

    @classmethod
    def _parse(cls, parsable):
        handshake_header_parser = cls._parse_handshake_header(parsable)

        parser = cls._parse_hello_header(handshake_header_parser['payload'])

        parser.parse_parsable('cipher_suite', TlsCipherSuiteFactory)
        parser.parse_parsable('compression_method', TlsCompressionMethodFactory)

        extension_parser = cls._parse_extensions(handshake_header_parser, parser, TlsExtensionsServer)

        return TlsHandshakeServerHello(
            protocol_version=parser['protocol_version'],
            random=parser['random'],
            session_id=parser['session_id'],
            compression_method=parser['compression_method'],
            cipher_suite=parser['cipher_suite'],
            extensions=parser['extensions'] if extension_parser else TlsExtensionsServer([]),
        ), handshake_header_parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.protocol_version)
        payload_composer.compose_parsable(self.random)
        payload_composer.compose_parsable(self.session_id)
        payload_composer.compose_numeric_enum_coded(self.cipher_suite)
        payload_composer.compose_numeric_enum_coded(self.compression_method)

        extension_bytes = self._compose_extensions(self.extensions)

        header_bytes = self._compose_header(payload_composer.composed_length + len(extension_bytes))

        return header_bytes + payload_composer.composed_bytes + extension_bytes


@attr.s
class TlsCertificate(ParsableBase):
    certificate = attr.ib(validator=attr.validators.instance_of(bytes))

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_bytes('certificate', 3)

        return TlsCertificate(bytes(parser['certificate'])), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_bytes(self.certificate, 3)

        return composer.composed_bytes


class TlsCertificates(VectorParsable):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(
            item_class=TlsCertificate,
            fallback_class=None,
            min_byte_num=1, max_byte_num=2 ** 24 - 1
        )


@attr.s
class TlsHandshakeCertificate(TlsHandshakeMessage):
    certificate_chain = attr.ib(validator=attr.validators.instance_of(TlsCertificates))

    @classmethod
    def get_handshake_type(cls):
        return TlsHandshakeType.CERTIFICATE

    @classmethod
    def _parse(cls, parsable):
        handshake_header_parser = cls._parse_handshake_header(parsable)

        parser = ParserBinary(handshake_header_parser['payload'])

        parser.parse_parsable('certificates', TlsCertificates)

        return TlsHandshakeCertificate(
            parser['certificates']
        ), handshake_header_parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()
        payload_composer.compose_parsable(self.certificate_chain)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


class TlsHandshakeCertificateStatus(TlsHandshakeMessage):
    def __init__(self, status_type, status):
        super().__init__()

        self.status_type = status_type
        self.status = status

    @classmethod
    def get_handshake_type(cls):
        return TlsHandshakeType.CERTIFICATE_STATUS

    @classmethod
    def _parse(cls, parsable):
        handshake_header_parser = cls._parse_handshake_header(parsable)

        parser = ParserBinary(handshake_header_parser['payload'])

        parser.parse_numeric('status_type', 1, TlsCertificateStatusType)
        parser.parse_bytes('status', 3)

        return TlsHandshakeCertificateStatus(
            parser['status_type'],
            parser['status']
        ), handshake_header_parser.parsed_length

    def compose(self):
        body_composer = ComposerBinary()
        body_composer.compose_numeric(self.status_type, 1)

        body_composer.compose_bytes(self.status, 3)

        header_bytes = self._compose_header(body_composer.composed_length)

        return header_bytes + body_composer.composed_bytes


class TlsHandshakeServerHelloDone(TlsHandshakeMessage):
    @classmethod
    def get_handshake_type(cls):
        return TlsHandshakeType.SERVER_HELLO_DONE

    @classmethod
    def _parse(cls, parsable):
        handshake_header_parser = cls._parse_handshake_header(parsable)

        if handshake_header_parser['payload']:
            raise InvalidValue(bytes(handshake_header_parser['payload']), TlsHandshakeServerHelloDone, 'payload')

        return TlsHandshakeServerHelloDone(), handshake_header_parser.parsed_length

    def compose(self):
        return self._compose_header(0)


class TlsECCurveType(enum.IntEnum):
    EXPLICIT_PRIME = 1
    EXPLICIT_CHAR2 = 2
    NAMED_CURVE = 3


@attr.s
class TlsHandshakeServerKeyExchange(TlsHandshakeMessage):
    param_bytes = attr.ib(validator=attr.validators.instance_of(bytes))

    @classmethod
    def get_handshake_type(cls):
        return TlsHandshakeType.SERVER_KEY_EXCHANGE

    @classmethod
    def _parse(cls, parsable):
        handshake_header_parser = cls._parse_handshake_header(parsable)

        return TlsHandshakeServerKeyExchange(
            bytes(handshake_header_parser['payload'])
        ), handshake_header_parser.parsed_length

    def compose(self):
        return self._compose_header(len(self.param_bytes)) + bytes(self.param_bytes)


class TlsClientCertificateType(enum.IntEnum):
    RSA_SIGN = 0x01
    DSS_SIGN = 0x02
    RSA_FIXED_DH = 0x03
    DSS_FIXED_DH = 0x04
    ECDSA_SIGN = 0x40
    RSA_FIXED_ECDH = 0x41
    ECDSA_FIXED_ECDH = 0x42
    GOST_SIGN256 = 0x43
    GOST_SIGN512 = 0x44


class TlsClientCertificateTypeVector(Vector):
    @classmethod
    def get_param(cls):
        return VectorParamNumeric(
            item_size=1,
            min_byte_num=1,
            max_byte_num=2 ** 8 - 1,
            numeric_class=TlsClientCertificateType
        )


class TlsDistinguishedName(Opaque):
    @classmethod
    def get_param(cls):
        return OpaqueParam(min_byte_num=1, max_byte_num=2 ** 16 - 1)


class TlsDistinguishedNameVector(VectorParsable):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(
            item_class=TlsDistinguishedName,
            fallback_class=None,
            min_byte_num=0, max_byte_num=2 ** 16 - 1
        )


@attr.s
class TlsHandshakeCertificateRequest(TlsHandshakeMessage):
    certificate_types = attr.ib(converter=TlsClientCertificateTypeVector)
    certificate_authorities = attr.ib(converter=TlsDistinguishedNameVector)
    supported_signature_algorithms = attr.ib(default=None)

    def __attrs_post_init__(self):
        if self.supported_signature_algorithms is not None:
            self.supported_signature_algorithms = TlsSignatureAndHashAlgorithmVector(
                self.supported_signature_algorithms
            )

        attr.validate(self)

    @classmethod
    def get_handshake_type(cls):
        return TlsHandshakeType.CERTIFICATE_REQUEST

    @classmethod
    def _parse(cls, parsable):
        handshake_header_parser = cls._parse_handshake_header(parsable)

        parser = ParserBinary(handshake_header_parser['payload'])

        parser.parse_parsable('certificate_types', TlsClientCertificateTypeVector)

        parser_remaining = ParserBinary(parser.unparsed)
        parser_remaining.parse_numeric('vector_length', 2)
        if parser_remaining['vector_length'] + 2 == parser.unparsed_length:
            supported_signature_algorithms = None
        else:
            parser.parse_parsable('supported_signature_algorithms', TlsSignatureAndHashAlgorithmVector)
            supported_signature_algorithms = parser['supported_signature_algorithms']

        parser.parse_parsable('certificate_authorities', TlsDistinguishedNameVector)

        msg = TlsHandshakeCertificateRequest(
            certificate_types=parser['certificate_types'],
            certificate_authorities=parser['certificate_authorities'],
            supported_signature_algorithms=supported_signature_algorithms,
        )

        return msg, handshake_header_parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.certificate_types)
        if self.supported_signature_algorithms is not None:
            payload_composer.compose_parsable(self.supported_signature_algorithms)
        payload_composer.compose_parsable(self.certificate_authorities)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


TLS_HANDSHAKE_HELLO_RETRY_REQUEST_RANDOM_BYTES = (
    b'\xcf\x21\xad\x74\xe5\x9a\x61\x11\xbe\x1d\x8c\x02\x1e\x65\xb8\x91' +
    b'\xc2\xa2\x11\x16\x7a\xbb\x8c\x5e\x07\x9e\x09\xe2\xc8\xa8\x33\x9c' +
    b''
)


TLS_HANDSHAKE_HELLO_RETRY_REQUEST_RANDOM = TlsHandshakeHelloRandom.parse_exact_size(
    TLS_HANDSHAKE_HELLO_RETRY_REQUEST_RANDOM_BYTES
)


@attr.s
class TlsHandshakeHelloRetryRequest(TlsHandshakeHello):
    cipher_suite = attr.ib(default=None, validator=attr.validators.in_(TlsCipherSuite))
    protocol_version = attr.ib(
        default=TlsProtocolVersion(TlsVersion.TLS1_3),
        validator=attr.validators.instance_of(TlsProtocolVersion),
    )
    random_bytes = attr.ib(
        default=TLS_HANDSHAKE_HELLO_RETRY_REQUEST_RANDOM,
        validator=attr.validators.instance_of(TlsHandshakeHelloRandom),
    )
    session_id = attr.ib(
        default=TlsSessionIdVector((random.randint(0, 255) for i in range(32))),
        converter=TlsSessionIdVector,
        validator=attr.validators.instance_of(TlsSessionIdVector),
    )
    compression_method = attr.ib(
        default=TlsCompressionMethod.NULL,
        converter=TlsCompressionMethod,
        validator=attr.validators.in_(TlsCompressionMethod),
    )
    extensions = attr.ib(
        default=TlsExtensionsServer([]),
        converter=TlsExtensionsServer,
        validator=attr.validators.instance_of(TlsExtensionsServer)
    )

    @classmethod
    def get_handshake_type(cls):
        return TlsHandshakeType.HELLO_RETRY_REQUEST

    @classmethod
    def _parse(cls, parsable):
        handshake_header_parser = cls._parse_handshake_header(parsable)
        parser = cls._parse_hello_header(handshake_header_parser['payload'])

        parser.parse_parsable('cipher_suite', TlsCipherSuiteFactory)

        parser.parse_parsable('compression_method', TlsCompressionMethodFactory)
        compression_method = parser['compression_method']
        session_id = parser['session_id']

        extension_parser = cls._parse_extensions(handshake_header_parser, parser, TlsExtensionsServer)

        return TlsHandshakeHelloRetryRequest(
            protocol_version=parser['protocol_version'],
            random_bytes=parser['random'],
            session_id=session_id,
            compression_method=compression_method,
            cipher_suite=parser['cipher_suite'],
            extensions=parser['extensions'] if extension_parser else TlsExtensionsClient([]),
        ), handshake_header_parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.protocol_version)
        payload_composer.compose_parsable(self.random_bytes)
        payload_composer.compose_parsable(self.session_id)
        payload_composer.compose_numeric_enum_coded(self.cipher_suite)
        payload_composer.compose_numeric_enum_coded(self.compression_method)

        extension_bytes = self._compose_extensions(self.extensions)

        header_bytes = self._compose_header(payload_composer.composed_length + len(extension_bytes))

        return header_bytes + payload_composer.composed_bytes + extension_bytes


class SslMessageBase(ParsableBase):
    @classmethod
    def get_message_type(cls):
        return NotImplementedError()  # pragma: no cover

    # pylint: disable=duplicate-code
    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    # pylint: disable=duplicate-code
    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class SslMessageType(enum.IntEnum):
    ERROR = 0x00
    CLIENT_HELLO = 0x01
    CLIENT_MASTER_KEY = 0x02
    CLIENT_FINISHED = 0x03
    SERVER_HELLO = 0x04
    SERVER_VERIFY = 0x05
    SERVER_FINISHED = 0x06
    REQUEST_CERTIFICATE = 0x07
    CLIENT_CERTIFICATE = 0x08


class SslCertificateType(enum.IntEnum):
    X509_CERTIFICATE = 0x01


class SslAuthenticationType(enum.IntEnum):
    MD5_WITH_RSA_ENCRYPTION = 0x01


class SslErrorType(enum.IntEnum):
    NO_CIPHER_ERROR = 0x0001
    NO_CERTIFICATE_ERROR = 0x0002
    BAD_CERTIFICATE_ERROR = 0x0003
    UNSUPPORTED_CERTIFICATE_TYPE_ERROR = 0x0004


@attr.s
class SslErrorMessage(SslMessageBase):
    error_type = attr.ib(validator=attr.validators.in_(SslErrorType))

    @classmethod
    def get_message_type(cls):
        return SslMessageType.ERROR

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_numeric('error_type', 2, SslErrorType)

        return SslErrorMessage(parser['error_type']), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.error_type, 2)

        return composer.composed_bytes


@attr.s
class SslHandshakeClientHello(SslMessageBase):
    cipher_kinds = attr.ib(validator=attr.validators.deep_iterable(member_validator=attr.validators.in_(SslCipherKind)))
    session_id = attr.ib(validator=attr.validators.instance_of(bytes))
    challenge = attr.ib(validator=attr.validators.instance_of(bytes))

    @session_id.default
    def _default_session_id(self):  # pylint: disable=no-self-use
        return bytes()

    @challenge.default
    def _default_challenge(self):  # pylint: disable=no-self-use
        return bytes(bytearray.fromhex(f'{random.getrandbits(128):16x}'.zfill(32)))

    @classmethod
    def get_message_type(cls):
        return SslMessageType.CLIENT_HELLO

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_parsable('version', TlsProtocolVersion)
        parser.parse_numeric('cipher_kinds_length', 2)
        parser.parse_numeric('session_id_length', 2)
        parser.parse_numeric('challenge_length', 2)
        parser.parse_parsable_array('cipher_kinds', parser['cipher_kinds_length'], SslCipherKindFactory)
        parser.parse_raw('session_id', parser['session_id_length'])
        parser.parse_raw('challenge', parser['challenge_length'])

        return SslHandshakeClientHello(
            cipher_kinds=parser['cipher_kinds'],
            session_id=bytes(parser['session_id']),
            challenge=bytes(parser['challenge']),
        ), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_parsable(TlsProtocolVersion(TlsVersion.SSL2))

        composer.compose_numeric(len(self.cipher_kinds) * 3, 2)
        composer.compose_numeric(len(self.session_id), 2)
        composer.compose_numeric(len(self.challenge), 2)
        composer.compose_numeric_array_enum_coded(self.cipher_kinds)
        composer.compose_raw(self.session_id)
        composer.compose_raw(self.challenge)

        return composer.composed_bytes


@attr.s
class SslHandshakeServerHello(SslMessageBase):
    certificate = attr.ib(validator=attr.validators.instance_of(bytes))
    cipher_kinds = attr.ib(validator=attr.validators.deep_iterable(member_validator=attr.validators.in_(SslCipherKind)))
    connection_id = attr.ib(validator=attr.validators.instance_of(bytes))
    session_id_hit = attr.ib(default=False, validator=attr.validators.instance_of(bool))

    @connection_id.default
    def _default_connection_id(self):  # pylint: disable=no-self-use
        return bytes()

    @classmethod
    def get_message_type(cls):
        return SslMessageType.SERVER_HELLO

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_numeric('session_id_hit', 1)
        parser.parse_numeric('certificate_type', 1, SslCertificateType)
        parser.parse_parsable('version', TlsProtocolVersion)
        parser.parse_numeric('certificate_length', 2)
        parser.parse_numeric('cipher_kinds_length', 2)
        parser.parse_numeric('connection_id_length', 2)
        parser.parse_raw('certificate', parser['certificate_length'])
        parser.parse_parsable_array('cipher_kinds', parser['cipher_kinds_length'], SslCipherKindFactory)
        parser.parse_raw('connection_id', parser['connection_id_length'])

        return SslHandshakeServerHello(
            certificate=bytes(parser['certificate']),
            cipher_kinds=parser['cipher_kinds'],
            connection_id=bytes(parser['connection_id']),
            session_id_hit=bool(parser['session_id_hit']),
        ), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(1 if self.session_id_hit else 0, 1)
        composer.compose_numeric(SslCertificateType.X509_CERTIFICATE, 1)
        composer.compose_parsable(TlsProtocolVersion(TlsVersion.SSL2))
        composer.compose_numeric(len(self.certificate), 2)
        composer.compose_numeric(len(self.cipher_kinds) * 3, 2)
        composer.compose_numeric(len(self.connection_id), 2)
        composer.compose_raw(self.certificate)
        composer.compose_numeric_array_enum_coded(self.cipher_kinds)
        composer.compose_raw(self.connection_id)

        return composer.composed_bytes


class TlsHandshakeMessageVariant(VariantParsable):
    _VARIANTS = collections.OrderedDict([
        (TlsHandshakeType.CLIENT_HELLO, [TlsHandshakeClientHello, ]),
        (TlsHandshakeType.SERVER_HELLO, [TlsHandshakeServerHello, ]),
        (TlsHandshakeType.CERTIFICATE, [TlsHandshakeCertificate, ]),
        (TlsHandshakeType.SERVER_KEY_EXCHANGE, [TlsHandshakeServerKeyExchange, ]),
        (TlsHandshakeType.CERTIFICATE_REQUEST, [TlsHandshakeCertificateRequest, ]),
        (TlsHandshakeType.CERTIFICATE_STATUS, [TlsHandshakeCertificateStatus, ]),
        (TlsHandshakeType.SERVER_HELLO_DONE, [TlsHandshakeServerHelloDone, ]),
        (TlsHandshakeType.HELLO_RETRY_REQUEST, [TlsHandshakeHelloRetryRequest, ]),
    ])

    @classmethod
    def _get_variants(cls):
        return cls._VARIANTS


class TlsSubprotocolMessageParser(SubprotocolParser):
    _SUBPROTOCOL_PARSERS = {
        TlsContentType.CHANGE_CIPHER_SPEC: TlsChangeCipherSpecMessage,
        TlsContentType.ALERT: TlsAlertMessage,
        TlsContentType.HANDSHAKE: TlsHandshakeMessageVariant,
        TlsContentType.APPLICATION_DATA: TlsApplicationDataMessage,
    }

    @classmethod
    def _get_subprotocol_parsers(cls):
        return cls._SUBPROTOCOL_PARSERS


class SslSubprotocolMessageParser(SubprotocolParser):
    _SUBPROTOCOL_PARSERS = {
        SslMessageType.ERROR: SslErrorMessage,
        SslMessageType.CLIENT_HELLO: SslHandshakeClientHello,
        SslMessageType.SERVER_HELLO: SslHandshakeServerHello,
    }

    @classmethod
    def _get_subprotocol_parsers(cls):
        return cls._SUBPROTOCOL_PARSERS
